import os

def value_str(value):
    ##convert values to string in namelist
    if isinstance(value, bool):
        if value:
            vstr = 'T'
        else:
            vstr = 'F'
    elif isinstance(value, str):
        vstr = "'"+value+"'"
    else:
        vstr = f'{value}'
    return vstr

def namelist(conf_dict, time, forecast_period, psi_init_type, member=0, dt_ratio=1, run_dir='.'):
    """Generate namelist for qg model
    Input:
    - m: Model class object with model configurations
    - dt_ratio: factor to multiply dt with
    """
    # model time step
    dt = conf_dict['dt'] * dt_ratio
    forecast_days = float(forecast_period) / 24
    dt_days = dt / conf_dict['tscale']
    # number of time steps to run
    total_counts = forecast_days / dt_days

    #random number based on member and time
    idum = 1000*member + 100*time.month + 10*time.day + time.hour

    #initialize energy only if at creation of initial condition
    #while during cycling, read the input from previous cycle and don't normalize energy
    if psi_init_type == 'read':
        initialize_energy = False
    else:
        initialize_energy = True

    ##start building the namelist content
    nmlstr = " &run_params\n"
    nmlstr += " kmax = "+value_str(conf_dict['kmax'])+"\n"
    nmlstr += " nz = "+value_str(conf_dict['nz'])+"\n"
    nmlstr += " F = "+value_str(conf_dict['F'])+"\n"
    nmlstr += " beta = "+value_str(conf_dict['beta'])+"\n"
    nmlstr += " adapt_dt = "+value_str(conf_dict['adapt_dt'])+"\n"
    nmlstr += " dt = "+value_str(dt)+"\n"
    nmlstr += " psi_init_file = 'input'\n"
    nmlstr += " psi_init_type = "+value_str(psi_init_type)+"\n"
    nmlstr += " initialize_energy = "+value_str(initialize_energy)+"\n"
    nmlstr += " e_o = "+value_str(conf_dict['e_o'])+"\n"
    nmlstr += " k_o = "+value_str(conf_dict['k_o'])+"\n"
    nmlstr += " delk = "+value_str(conf_dict['delk'])+"\n"
    nmlstr += " m_o = "+value_str(conf_dict['m_o'])+"\n"
    nmlstr += " z_o = "+value_str(conf_dict['z_o'])+"\n"
    nmlstr += " strat_type = "+value_str(conf_dict['strat_type'])+"\n"
    nmlstr += " deltc = "+value_str(conf_dict['deltc'])+"\n"
    nmlstr += " ubar_type = "+value_str(conf_dict['ubar_type'])+"\n"
    nmlstr += " delu = "+value_str(conf_dict['delu'])+"\n"
    nmlstr += " uscale = "+value_str(conf_dict['uscale'])+"\n"
    nmlstr += " use_topo = "+value_str(conf_dict['use_topo'])+"\n"
    nmlstr += " topo_type = "+value_str(conf_dict['topo_type'])+"\n"
    nmlstr += " k_o_topo = "+value_str(conf_dict['k_o_topo'])+"\n"
    nmlstr += " del_topo = "+value_str(conf_dict['del_topo'])+"\n"
    nmlstr += " use_forcing = "+value_str(conf_dict['use_forcing'])+"\n"
    nmlstr += " norm_forcing = "+value_str(conf_dict['norm_forcing'])+"\n"
    nmlstr += " forc_coef = "+value_str(conf_dict['forc_coef'])+"\n"
    nmlstr += " forc_corr = "+value_str(conf_dict['forc_corr'])+"\n"
    nmlstr += " kf_min = "+value_str(conf_dict['kf_min'])+"\n"
    nmlstr += " kf_max = "+value_str(conf_dict['kf_max'])+"\n"
    nmlstr += " filter_type = "+value_str(conf_dict['filter_type'])+"\n"
    nmlstr += " filter_exp = "+value_str(conf_dict['filter_exp'])+"\n"
    nmlstr += " filt_tune = "+value_str(conf_dict['filt_tune'])+"\n"
    nmlstr += " k_cut = "+value_str(conf_dict['k_cut'])+"\n"
    nmlstr += " rho_slope = "+value_str(conf_dict['rho_slope'])+"\n"
    nmlstr += " bot_drag = "+value_str(conf_dict['bot_drag'])+"\n"
    nmlstr += " therm_drag = "+value_str(conf_dict['therm_drag'])+"\n"
    nmlstr += " idum = "+value_str(idum)+"\n"
    nmlstr += " total_counts = "+value_str(total_counts)+"\n"
    nmlstr += " write_step = "+value_str(total_counts)+"\n"
    nmlstr += " diag1_step = "+value_str(total_counts)+"\n"
    nmlstr += " diag2_step = "+value_str(total_counts)+"\n"
    nmlstr += " /"

    ##write the namelist to input.nml file
    with open(os.path.join(run_dir, 'input.nml'), 'wt') as f:
        f.write(nmlstr)




import numpy as np
import unittest
from pyproj import Proj
from NEDAS.grid import Grid

class TestGrid(unittest.TestCase):
    def test_regular_grid_creation(self):
        proj = Proj('+proj=stere')
        xstart, xend, ystart, yend = 0, 100, 0, 100
        dx = 1
        grid = Grid.regular_grid(proj, xstart, xend, ystart, yend, dx)
        self.assertEqual(grid.nx, 100)
        self.assertEqual(grid.ny, 100)
        self.assertAlmostEqual(grid.dx, 1.0)
        self.assertAlmostEqual(grid.dy, 1.0)

    def test_random_grid_creation(self):
        proj = Proj('+proj=stere')
        xstart, xend, ystart, yend = 0, 100, 0, 100
        npoints = 1000
        grid = Grid.random_grid(proj, xstart, xend, ystart, yend, npoints)
        self.assertEqual(grid.x.size, npoints)

    def test_find_index_regular(self):
        proj = Proj('+proj=stere')
        xstart, xend, ystart, yend = 0, 100, 0, 100
        dx = 1
        grid = Grid.regular_grid(proj, xstart, xend, ystart, yend, dx)

        ##case 1: find one of the grid point
        x, y = 10, 10
        inside, _, vertices, in_coords, nearest = grid.find_index(x, y)
        self.assertTrue(inside[0])
        self.assertTrue((vertices[0] == np.array([1010, 1011, 1111, 1110])).all())
        self.assertTrue((in_coords[0] == np.array([0., 0.])).all())
        self.assertEqual(nearest[0], 1010)

        ##case 2: find a point inside the grid
        x, y = 10.8, 10.2
        inside, _, vertices, in_coords, nearest = grid.find_index(x, y)
        self.assertTrue(inside[0])
        self.assertTrue((vertices[0] == np.array([1010, 1011, 1111, 1110])).all())
        self.assertAlmostEqual(in_coords[0,0], 0.8)
        self.assertAlmostEqual(in_coords[0,1], 0.2)
        self.assertEqual(nearest[0], 1011)

        ##case 3: find a point outside the grid
        x, y = -1, -1
        inside, _, _, _, _ = grid.find_index(x, y)
        self.assertFalse(inside[0])

    def test_find_index_irregular(self):
        proj = Proj('+proj=stere')
        x, y = np.meshgrid(np.arange(10), np.arange(10))
        grid = Grid(proj, x.flatten(), y.flatten(), regular=False)

        ##case 1: find a point inside the irregular mesh
        x, y = 1.2, 2.7
        inside, indices, vertices, in_coords, nearest = grid.find_index(x, y)
        self.assertTrue(inside[0])
        self.assertEqual(indices[0], 22)
        self.assertTrue((vertices[0] == np.array([31, 21, 32])).all())
        self.assertAlmostEqual(in_coords[0,0], 0.5)
        self.assertAlmostEqual(in_coords[0,1], 0.3)
        self.assertAlmostEqual(in_coords[0,2], 0.2)
        self.assertEqual(nearest[0], 31)

    def test_cyclic_dim(self):
        proj = Proj('+proj=stere')
        xstart, xend, ystart, yend = 0, 100, 0, 100
        dx = 1
        grid = Grid.regular_grid(proj, xstart, xend, ystart, yend, dx, cyclic_dim='xy')

        ##a point outside grid boundary will still be found "inside"
        x, y = -1, -1
        inside, _, vertices, in_coords, nearest = grid.find_index(x, y)
        self.assertTrue(inside[0])
        self.assertTrue((vertices[0] == np.array([9999, 9900, 0, 99])).all())
        self.assertTrue((in_coords[0] == np.array([0., 0.])).all())
        self.assertEqual(nearest[0], 9999)

    def test_pole_dim(self):
        grid1 = Grid.regular_grid(Proj('+proj=longlat'), -180, 181, 0, 91, 1, cyclic_dim='x')  ##lon lat grid
        grid2 = Grid.regular_grid(Proj('+proj=stere +lat_0=90'), -1e6, 1e6, -1e6, 1e6, 1e4, centered=True)  ##polar stereographic grid

        ##a vector field with all zonal winds
        u = np.ones(grid1.x.shape)
        v = np.zeros(grid1.x.shape)
        vfld1 = np.array([u, v])

        ##case 1: pole_dim is not set, there will be nan after rotating vectors
        grid1.set_destination_grid(grid2)
        vfld2 = grid1.convert(vfld1, is_vector=True)
        self.assertTrue(np.isnan(vfld2[0, 100, 100]))

        ##case 2: when pole_dim is set, the void will be filled, so nan is gone
        grid1.pole_dim='y'
        grid1.pole_index=(-1,)
        grid1.set_destination_grid(grid2)
        vfld2 = grid1.convert(vfld1, is_vector=True)
        self.assertFalse(np.isnan(vfld2).any())

    def test_rotate_vector(self):
        grid1 = Grid.regular_grid(Proj('+proj=stere +lat_0=90 +lon_0=0'), -1e6, 1e6, -1e6, 1e6, 1e4)
        grid2 = Grid.regular_grid(Proj('+proj=stere +lat_0=90 +lon_0=90'), -1e6, 1e6, -1e6, 1e6, 1e4)
        u = np.ones(grid1.x.shape)
        v = np.zeros(grid1.x.shape)
        vfld1 = np.array([u, v])

        ##at domain center, vfld1 is (1, 0), after rotating to grid2, vfld2 should be (0, -1)
        grid1.set_destination_grid(grid2)
        vfld2 = grid1.convert(vfld1, is_vector=True)
        self.assertAlmostEqual(vfld2[0, 100, 100], 0.0)
        self.assertAlmostEqual(vfld2[1, 100, 100], -1.0)

if __name__ == '__main__':
    unittest.main()




import numpy as np
from NEDAS.utils.progress import progress_bar
from NEDAS.assim_tools.inflation.base import Inflation

class RTPPInflation(Inflation):
    def adaptive_prior_inflation(self, c, state, obs):
        raise NotImplementedError("Relaxation method is only implemented for posterior ensemble")

    def adaptive_post_inflation(self, c, state, obs):
        """Adaptive covariance relaxation method (Ying and Zhang 2015, QJRMS)"""
        c.print_1p(">>> adaptive covariance relaxation:\n")
        stats = self.obs_space_stats(c, state, obs)
        if stats['total_nobs'] < 3:
            c.print_1p(f"Warning: insufficient nobs to establish statistics, setting self.coef=0")
            self.coef = 0.
        else:
            varb = stats['varb'] / stats['total_nobs']
            vara = stats['vara'] / stats['total_nobs']
            varo = stats['varo'] / stats['total_nobs']
            omb2 = stats['omb2'] / stats['total_nobs']
            omaamb = stats['omaamb'] / stats['total_nobs']
            amb2 = stats['amb2'] / stats['total_nobs']
            beta = np.sqrt(varb/vara)
            lamb = np.sqrt(max(0.0, (omb2-varo-amb2)/vara))
            c.print_1p(f"varb = {varb}, vara = {vara}, varo={varo}\n")
            c.print_1p(f"omb2 = {omb2}, omaamb = {omaamb}, amb2={amb2}\n")
            c.print_1p(f"beta = {beta}, lambda = {lamb}\n")
            if beta <= 1:
                self.coef = 0
            else:
                self.coef = (lamb - 1) / (beta - 1)
            if self.coef > 2:
                self.coef = 2
            if self.coef < -1:
                self.coef = -1

    def apply_inflation(self, c, state, flag):
        pid_mem_show = [p for p,lst in state.mem_list.items() if len(lst)>0][0]
        pid_rec_show = [p for p,lst in state.rec_list.items() if len(lst)>0][0]
        c.pid_show = pid_rec_show * c.nproc_mem + pid_mem_show
        c.print_1p(f'relaxing to prior ensemble perturbations with coef={self.coef}\n')

        ##process the fields, each processor goes through its own subset of
        ##mem_id,rec_id simultaneously
        nm = len(state.mem_list[c.pid_mem])
        nr = len(state.rec_list[c.pid_rec])

        for r, rec_id in enumerate(state.rec_list[c.pid_rec]):
            ##read the mean field with rec_id
            fld_prior_mean = state.read_field(state.prior_mean_file, c.grid.mask, 0, rec_id)
            fld_post_mean = state.read_field(state.post_mean_file, c.grid.mask, 0, rec_id)

            for m, mem_id in enumerate(state.mem_list[c.pid_mem]):
                if c.debug:
                    print(f"PID {c.pid:4}: relax_to_prior_perturb mem{mem_id+1:03}", flush=True)
                else:
                    c.print_1p(progress_bar(m*nr+r, nm*nr))
                ##inflate the ensemble perturbations by relaxing to prior perturbations
                fld_prior = state.fields_prior[mem_id, rec_id]
                fld_post = state.fields_post[mem_id, rec_id]
                fld_post = fld_post_mean + self.coef*(fld_prior - fld_prior_mean) + (1.-self.coef)*(fld_post - fld_post_mean)

        c.comm.Barrier()
        c.print_1p(' done.\n')



import numpy as np
from NEDAS.utils.random_perturb import random_field_powerlaw
from NEDAS.utils.fft_lib import fft2, ifft2, get_wn

def initial_condition(grid, Vmax, Rmw, Vbg, Vslope, loc_sprd=0):
    """
    Initialize the 2d vortex model with a Rankine vortex embedded in a random wind flow.

    Args:
        grid (Grid): The model domain, doubly periodic, described by a Grid obj
        Vmax (float): Maximum wind speed (vortex intensity), m/s
        Rmw (float): Radius of maximum wind (vortex size), m
        Vbg (float): Background flow average wind speed, m/s
        Vslope (int): Background flow kinetic energy spectrum power law (typically -2)
        loc_sprd (float, optional): The ensemble spread in vortex center position, m

    Returns:
        np.ndarray: The vector velocity field, shape (2, ny, nx).
    """
    ##the vortex is randomly placed in the domain
    center_x = 0.5*(grid.xmin+grid.xmax) + np.random.normal(0, loc_sprd)
    center_y = 0.5*(grid.ymin+grid.ymax) + np.random.normal(0, loc_sprd)

    vortex = rankine_vortex(grid, Vmax, Rmw, center_x, center_y)

    ##the background wind field is randomly drawn
    bkg_flow = random_flow(grid, Vbg, Vslope)

    return vortex + bkg_flow

def rankine_vortex(grid, Vmax, Rmw, center_x, center_y):
    """
    Generate a Rankine vortex velocity field.

    Args:
        grid (Grid): The model domain, doubly periodic
        Vmax (float): Maximum wind speed (vortex intensity), m/s
        Rmw (float): Radius of maximum wind (vortex size), m
        center_x (float): Vortex center X-coordinate.
        center_y (float): Vortex center Y-coordinate.

    Returns:
        np.ndarray: The vector velocity field with shape (2, ny, nx)
    """
    ##radius from vortex center
    r = np.hypot(grid.x - center_x, grid.y - center_y)
    r[np.where(r==0)] = 1e-10  ##avoid divide by 0

    ##wind speed profile with radius
    wspd = np.zeros(r.shape)
    ind = np.where(r <= Rmw)
    wspd[ind] = Vmax * r[ind] / Rmw
    ind = np.where(r > Rmw)
    wspd[ind] = Vmax * (Rmw / r[ind])**1.5
    wspd[np.where(r==0)] = 0

    u = -wspd * (grid.y - center_y) / r
    v = wspd * (grid.x - center_x) / r

    return np.array([u, v])

def random_flow(grid, amp, power_law):
    """
    Generate a random velocity field as the background flow

    Args:
        grid (Grid): The model domain, doubly periodic, described by a Grid obj
        amp (float): wind speed amplitude, m/s
        power_law (int): wind kinetic energy spectrum power law (typically -2)

    Returns:
        np.ndarray: The vector velocity field with shape (2, ny, nx)
    """
    ny, nx = grid.x.shape
    fld = np.zeros((2, ny, nx))
    dx = grid.dx

    ##generate random streamfunction for the wind
    ##note: streamfunc powerlaw = wind powerlaw - 2
    psi = random_field_powerlaw(nx, ny, 1, power_law-2)

    ##convert to wind
    u = -(np.roll(psi, -1, axis=0) - np.roll(psi, 1, axis=0)) / (2.0*dx)
    v = (np.roll(psi, -1, axis=1) - np.roll(psi, 1, axis=1)) / (2.0*dx)

    ##normalize and scale to the required wind amp
    u = amp * (u - np.mean(u)) / np.std(u)
    v = amp * (v - np.mean(v)) / np.std(v)

    return np.array([u, v])

def advance_time(fld, dx, t_intv, dt, gen, diss):
    """
    Advance forward in time to integrate the model (forecasting)

    Args:
        fld (np.ndarray): The prognostic velocity field with shape (2,ny,nx)
        dx (float): Model grid spacing, meter
        t_intv (float): Integration time period, hour
        dt (float): Model time step, second
        gen (float): Vorticity generation rate
        diss (float): Dissipation rate

    Returns:
        np.ndarray: The forecast velocity field
    """
    ##input wind components, convert to spectral space
    uh = fft2(fld[0, :, :])
    vh = fft2(fld[1, :, :])

    ##convert to zeta
    ki, kj = get_scaled_wn(uh, dx)
    zetah = 1j * (ki*vh - kj*uh)
    k2 = ki**2 + kj**2
    k2[0, 0] = 1.
    #k2 = np.where(k2!=0, k2, np.ones_like(k2)) #avoid singularity in inversion

    ##run time loop:
    ##t_intv is run period in hours
    ##dt is model time step in seconds
    for n in range(int(t_intv*3600/dt)):
        ##use rk4 numeric scheme to integrate forward in time:
        rhs1 = forcing(uh, vh, zetah, dx, gen, diss)
        zetah1 = zetah + 0.5*dt*rhs1
        rhs2 = forcing(uh, vh, zetah1, dx, gen, diss)
        zetah2 = zetah + 0.5*dt*rhs2
        rhs3 = forcing(uh, vh, zetah2, dx, gen, diss)
        zetah3 = zetah + dt*rhs3
        rhs4 = forcing(uh, vh, zetah3, dx, gen, diss)
        zetah = zetah + dt*(rhs1/6.0 + rhs2/3.0 + rhs3/3.0 + rhs4/6.0)

        ##inverse zeta to get u, v
        psih = -zetah / k2
        uh = -1j * kj * psih
        vh = 1j * ki * psih

    u = ifft2(uh)
    v = ifft2(vh)

    return np.array([u, v])

def get_scaled_wn(x, dx):
    """scaled wavenumber k for pseudospectral method"""
    n = x.shape[0]
    wni, wnj = get_wn(x)
    ki = (2.*np.pi) * wni / (n*dx)
    kj = (2.*np.pi) * wnj / (n*dx)
    return ki, kj

def forcing(u, v, zeta, dx, gen, diss):
    """forcing terms on RHS of prognostic equations"""
    ki, kj = get_scaled_wn(zeta, dx)
    ug = ifft2(u)
    vg = ifft2(v)
    ##advection term:
    f = -fft2(ug*ifft2(1j*ki*zeta) + vg*ifft2(1j*kj*zeta))
    ##generation term:
    vmax = np.max(np.sqrt(ug**2+vg**2))
    if vmax > 75:  ##cut off generation if vortex intensity exceeds limit
        gen = 0
    n = zeta.shape[0]
    k2d = np.sqrt(ki**2 + kj**2)*(n*dx)/(2.*np.pi)
    kc = 8
    dk = 3
    gen_response = np.exp(-0.5*(k2d-kc)**2/dk**2)
    f += gen*gen_response*zeta
    ##dissipation term:
    f -= diss*(ki**2+kj**2)*zeta
    return f




import numpy
import sys
import os
import re
import argparse
import yaml

def convert_notation(data):
    """
    Parse values in data, convert strings to appropriate types if possible.

    Goes through the data recursively if it is a list or a dictionary. If the string can be interpreted as a
    scientific notation, inf, or boolean flag, convert it to the appropriate type.

    Args:
        data (any): The data to be converted.

    Returns:
        any: The converted data.
    """
    if isinstance(data, dict):
        return {key: convert_notation(value) for key, value in data.items()}
    elif isinstance(data, list):
        return [convert_notation(element) for element in data]
    elif isinstance(data, str):
        ## Match scientific notation pattern and convert to float
        if re.match(r'^-?\d+(\.\d*)?[eE]-?\d+$', data):
            return float(data)
        ## convert "inf" to numpy.inf
        elif data.lower() == "inf":
            return numpy.inf
        elif data.lower() == "-inf":
            return -numpy.inf
        ## convert string to bool
        elif data.lower() in ['y', 'yes', 'on', 't', 'true', '.true.']:
            return True
        elif data.lower() in ['n', 'no', 'off', 'f', 'false', '.false.']:
            return False
        else:
            return data
    else:
        return data

def str2bool(data:str) -> int:
    """
    Convert a string to a boolean value (0 or 1).
    """
    if data.lower() in ['y', 'yes', 'on', 't', 'true', '.true.']:
        return 1
    elif data.lower() in ['n', 'no', 'off', 'f', 'false', '.false.']:
        return 0
    else:
        raise ValueError(f"Invalid input value for str2bool: {data}")

def parse_config(code_dir='.', config_file=None, parse_args=False, **kwargs):
    """
    Load configuration from YAML files and runtime arguments.

    This function loads configuration settings from a default YAML file, located in `code_dir/default.yml`.
    If `config_file` is provided, then values defined in that file will overwrite those in the default YAML file.
    If additional key=value pairs are provided, either through runtime command-line arguments or as `kwargs`,
    those will also overwrite the existing values.

    The default YAML file also serves as a template for the ArgumentParser, which parses the key-value pairs,
    if the value is not None, parser will add the argument with type and default value in the help message.

    Args:
        code_dir (str, optional): Directory containing the `default.yml` file. Default is the current directory.
        config_file (str, optional): Alternative YAML config file to overwrite default settings.
        parse_args (bool, optional): If True, parse runtime arguments with argparse. Default is False.
        **kwargs: Additional configuration key-value pairs that can overwrite existing settings.

    Returns:
        dict: A dictionary containing all configuration variables.
    """
    if parse_args:
        input_args = sys.argv[1:]
        if len(input_args) == 0:
            print(f"Usage: {sys.argv[0]} -c YAML_config_file")
            exit()
    else:
        input_args = {}

    default_config_file = os.path.join(code_dir, 'default.yml')
    if os.path.exists(default_config_file):
        with open(default_config_file, 'r') as f:
            config_dict = yaml.safe_load(f)
    else:
        config_dict = {}

    ##optionally, a config file can be specified at runtime
    ##through the config_file argument, build a parser for this
    parser = argparse.ArgumentParser(description='Parse configuration', add_help=False)
    parser.add_argument('-c', '--config_file')
    parser.add_argument('-h', '--help', action='store_true')

    ##parse --config_file and --help first
    args, remaining_args = parser.parse_known_args(input_args) # type: ignore

    ##update config_dict if new config_file is provided
    if config_file is not None:
        with open(config_file, 'r') as f:
            config_dict = {**config_dict, **yaml.safe_load(f)}

    if args.config_file is not None:
        with open(args.config_file, 'r') as f:
            config_dict = {**config_dict, **yaml.safe_load(f)}

    if not isinstance(config_dict, dict):
        config_dict = {}

    ##append new config variables defined in kwargs
    config_dict = {**config_dict, **kwargs}

    ##continue building the parser with additional arguments to update
    ##individual config variables in config_dict
    parser = argparse.ArgumentParser()

    for key, value in config_dict.items():
        key_rec = {}
        value = convert_notation(value)
        key_rec['default'] = value
        if value is not None:
            ##bool variable type needs special treatment for parsing runtime input string
            if isinstance(value, bool):
                key_rec['type'] = lambda x: bool(str2bool(x))
            else:
                key_rec['type'] = type(value)
            ##help message shows the default value and type for this argument
            key_rec['help'] = f"type: {type(value).__name__}, default: {value}".replace('%','%%')

        ##add the argument to parser
        parser.add_argument('--'+key, **key_rec)

    ##show help message
    if args.help:
        print(f"""
Default configuration variables are defined in 'default.yml' in the code directory.

You can specify a YAML config file by [-c YAML_FILE] or [--config_file YAML_FILE] to overwrite the default configuration.

Furthermore, you can also overwrite some configuration variables by specifying them at runtime:
""")
        parser.print_help()
        exit()

    ##run the parser to get the config namespace object
    config, remaining_args = parser.parse_known_args(remaining_args)

    ##return the dict with config variables
    return vars(config)


"""Diagnostic module to plot the observation sequence and obs_prior ensemble"""

import os
import numpy as np
import matplotlib.pyplot as plt
from NEDAS.utils.conversion import ensure_list, dt1h
from NEDAS.utils.shell_utils import makedir
from NEDAS.utils.graphics import add_colorbar, adjust_ax_size, get_cmap
from NEDAS.assim_tools.state import State
from NEDAS.assim_tools.obs import Obs

def get_task_list(c, **kwargs) -> list:

    variables = ensure_list(kwargs['variables'])
    dataset_src = ensure_list(kwargs['dataset_src'])
    vmin = ensure_list(kwargs['vmin'])
    vmax = ensure_list(kwargs['vmax'])
    nlevels = ensure_list(kwargs['nlevels'])
    cmap = ensure_list(kwargs['cmap'])
    vmin_diff = ensure_list(kwargs['vmin_diff'])
    vmax_diff = ensure_list(kwargs['vmax_diff'])
    nlevels_diff = ensure_list(kwargs['nlevels_diff'])
    cmap_diff = ensure_list(kwargs['cmap_diff'])

    state = State(c)
    obs = Obs(c, state)

    ##observation time steps within window
    obs_window_min = kwargs.get('obs_window_min', 0)
    obs_window_max = kwargs.get('obs_window_max', 0)
    obs_dt = ensure_list(kwargs['obs_dt'])
    obs_kmin = ensure_list(kwargs['obs_kmin'])
    obs_kmax = ensure_list(kwargs['obs_kmax'])

    tasks = []
    for i, vname in enumerate(variables):
        ##check if obs rec is defined in obs.info
        obs_rec_query = [id for id,r in obs.info['records'].items() if r['name']==vname and r['dataset_src']==dataset_src[i]]
        assert len(obs_rec_query)>0, f"cannot find obs record for '{vname}' from dataset '{dataset_src[i]}'"
        obs_rec_id = obs_rec_query[0]

        ##time steps for this obs
        obs_ts = c.time + np.arange(obs_window_min, obs_window_max, obs_dt[i]) * dt1h

        ##vertical levels for this obs
        levels = np.arange(obs_kmin[i], obs_kmax[i]+1)

        for k in levels:
            for t in obs_ts:
                for m in range(c.nens):
                    tasks.append({**kwargs, 'obs_rec_id':obs_rec_id, 'member':m, 'k':k, 't':t, 'dt':obs_dt[i], 'vmin':vmin[i], 'vmax':vmax[i], 'nlevels':nlevels[i], 'cmap':cmap[i], 'vmin_diff':vmin_diff[i], 'vmax_diff':vmax_diff[i], 'nlevels_diff':nlevels_diff[i], 'cmap_diff':cmap_diff[i]})
    return tasks

def run(c, **kwargs) -> None:
    """
    Run diagnostics: plot the ensemble states
    """
    if 'plot_dir' in kwargs:
        plot_dir = kwargs['plot_dir']
    else:
        plot_dir = os.path.join(c.work_dir, 'plots', 'observations')
    makedir(plot_dir)

    state = State(c)
    obs = Obs(c, state)

    figsize = (kwargs.get('fig_size_x', 16), kwargs.get('fig_size_y', 7))
    landcolor = kwargs.get('land_color', 'gray')

    obs_rec_id = kwargs['obs_rec_id']
    member = kwargs['member']
    obs_rec = obs.info['records'][obs_rec_id]
    vmin = kwargs['vmin']
    vmax = kwargs['vmax']
    nlevels = kwargs['nlevels']
    cmap = get_cmap(kwargs['cmap'])
    vmin_diff = kwargs['vmin_diff']
    vmax_diff = kwargs['vmax_diff']
    nlevels_diff = kwargs['nlevels_diff']
    cmap_diff = get_cmap(kwargs['cmap_diff'])

    k = kwargs['k']
    t = kwargs['t']
    dt = kwargs['dt']
    if c.debug:
        print(f"PID {c.pid:4} plotting observations '{obs_rec['name']:20}' from {obs_rec['dataset_src']} at level {k:3} {t} ~ {t+dt*dt1h}", flush=True)

    ##if the viewer html file does not exist, generate it
    viewer = os.path.join(plot_dir, 'index.html')
    if not os.path.exists(viewer):
        generate_viewer_html(c, plot_dir, figsize, **kwargs)

    ##plot the variables defined in kwargs, save to figfile
    figfile = os.path.join(plot_dir, f"{obs_rec['dataset_src']}_{obs_rec['name']}_k{k}_{t:%Y%m%dT%H%M%S}_{t+dt*dt1h:%Y%m%dT%H%M%S}_mem{member+1:03}.png")

    ##read the obs data from analysis_dir/obs_seq
    obs_seq = np.load(os.path.join(state.analysis_dir, f'obs_seq.rec{obs_rec_id}.npy'), allow_pickle=True).item()
    obs_prior_seq = np.load(os.path.join(state.analysis_dir, f'obs_prior_seq.rec{obs_rec_id}.mem{member:03}.npy'), allow_pickle=True)

    ##filter for the obs within time and vertical level range
    tmask = (obs_seq['t'] > t) & (obs_seq['t'] <= t+dt*dt1h)
    obs_z = np.abs(obs_seq['z'])
    obs_x = obs_seq['x']
    obs_y = obs_seq['y']
    model_z = np.abs(obs.read_mean_z_coords(c, state, c.time))
    if k == 0:
        zk = c.grid.interp(model_z[k], obs_x, obs_y)
        zmask = (obs_seq['z'] == zk)
    else:
        zk = c.grid.interp(model_z[k], obs_x, obs_y)
        zk1 = c.grid.interp(model_z[k-1], obs_x, obs_y)
        zmask = (obs_z > zk1) & (obs_z <= zk)
    ind = np.where(tmask & zmask)[0]

    ##plot the observations as scattered data over c.grid
    try:
        fig, ax = plt.subplots(1, 2, figsize=figsize)
        if obs_rec['is_vector']:
            obs_u = obs_seq['obs'][0,...][ind]
            obs_v = obs_seq['obs'][1,...][ind]
            obs = np.array([obs_u, obs_v])
            c.grid.plot_scatter(ax[0], obs, vmin, vmax, nlevels, x=obs_x[ind], y=obs_y[ind], is_vector=True, units=obs_rec['units'])
            adjust_ax_size(ax[0])

            obs_prior_u = obs_prior_seq[0,...][ind]
            obs_prior_v = obs_prior_seq[1,...][ind]
            obs_prior = np.array([obs_prior_u, obs_prior_v])
            obs_diff = obs - obs_prior
            c.grid.plot_scatter(ax[1], obs_diff, vmin_diff, vmax_diff, nlevels_diff, x=obs_x[ind], y=obs_y[ind], is_vector=True, units=obs_rec['units'])
            adjust_ax_size(ax[1])

        else:
            obs = obs_seq['obs'][ind]
            c.grid.plot_scatter(ax[0], obs, vmin, vmax, nlevels, cmap=cmap, markersize=10, x=obs_x[ind], y=obs_y[ind])
            add_colorbar(fig, ax[0], cmap, vmin, vmax, nlevels, units=obs_rec['units'])

            obs_prior = obs_prior_seq[ind]
            obs_diff = obs - obs_prior
            c.grid.plot_scatter(ax[1], obs_diff, vmin_diff, vmax_diff, nlevels_diff, cmap=cmap_diff, markersize=10, x=obs_x[ind], y=obs_y[ind])
            add_colorbar(fig, ax[1], cmap_diff, vmin_diff, vmax_diff, nlevels_diff, units=obs_rec['units'])

        for i in range(2):
            c.grid.plot_land(ax[i], color=landcolor)
            ax[i].set_xlabel('x (m)', fontsize=14)
            ax[i].set_ylabel('y (m)', fontsize=14)

        ax[0].set_title(f"Observation", fontsize=14)
        ax[1].set_title(f"Diff(Observation - Model member {member+1})", fontsize=14)
        fig.suptitle(f"{obs_rec['dataset_src']}_{obs_rec['name']}, level {k}, {t} ~ {t+dt*dt1h}", fontsize=16)
        plt.savefig(figfile)
        plt.close()

    except Exception as e:
        print(f"ERROR: Failed to plot {obs_rec['name']} at level {k} and time {t} ~ {t+dt*dt1h}")
        raise e

def generate_viewer_html(c, plot_dir, figsize, **kwargs) -> None:
    """Generating a html page to help viewing the ensemble state variables"""
    if c.debug:
        print(f"Generating viewer.html page in {plot_dir}")

    with open(os.path.join(os.path.dirname(__file__), 'viewer.html'), 'rt') as f:
        html_page = f.read()

    variables = ensure_list(kwargs['variables'])
    dataset_src = ensure_list(kwargs['dataset_src'])
    obs_window_min = kwargs.get('obs_window_min', 0)
    obs_window_max = kwargs.get('obs_window_max', 0)
    obs_dt = ensure_list(kwargs['obs_dt'])
    obs_kmin = ensure_list(kwargs['obs_kmin'])
    obs_kmax = ensure_list(kwargs['obs_kmax'])

    levels_by_variable = ""
    times_by_variable = ""
    for i, vname in enumerate(variables):
        name = f"{dataset_src[i]}_{vname}"
        obs_ts = c.time + np.arange(obs_window_min, obs_window_max, obs_dt[i]) * dt1h
        levels = np.arange(obs_kmin[i], obs_kmax[i]+1)

        levels_by_variable += f"'{name}': ["
        for level in levels:
            levels_by_variable += f"{level}, "
        levels_by_variable += "], \n"

        times_by_variable += f"'{name}': ["
        for t in obs_ts:
            times_by_variable += f"'{t:%Y%m%dT%H%M%S}_{t+obs_dt[i]*dt1h:%Y%m%dT%H%M%S}', "
        times_by_variable += "], \n"
    html_page = html_page.replace("LEVELS_BY_VARIABLE", levels_by_variable)
    html_page = html_page.replace("TIMES_BY_VARIABLE", times_by_variable)

    members = "["
    for m in range(c.nens):
        members += f"'{m+1:03}', "
    members += "]"
    html_page = html_page.replace("MEMBERS", members)

    html_page = html_page.replace("TITLE", "Observations")
    html_page = html_page.replace("IMAGE_WIDTH", f"{figsize[0]*60}")
    html_page = html_page.replace("IMAGE_HEIGHT", f"{figsize[1]*60}")

    ##write the html page to file
    with open(os.path.join(plot_dir, 'index.html'), 'w') as f:
        f.write(html_page)



import os
import numpy as np
from NEDAS.utils.conversion import t2h, h2t, dt1h, ensure_list
from NEDAS.utils.progress import progress_bar
from NEDAS.utils.parallel import bcast_by_root, distribute_tasks

class Obs:
    """
    Class for handling observations.
    
    The observation has dimensions: variable, time, z, y, x
    Since the observation network is typically irregular, we store the obs record
    for each variable in a 1d sequence, with coordinates (t,z,y,x), and size nobs

    To parallelize workload, we distribute each obs record over all the processors

    for batch assimilation mode, each pid stores the list of local obs within the
    hroi of its tiles, with size nlobs (number of local obs)
    
    for serial mode, each pid stores a non-overlapping subset of the obs list,
    here 'local' obs (in storage sense) is broadcast to all pid before computing
    its update to the state/obs near that obs.

    The hroi is separately defined for each obs record.
    For very large hroi, the serial mode is more parallel efficient option, since
    in batch mode the same obs may need to be stored in multiple pids

    To compare to the observation, obs_prior simulated by the model needs to be
    computed, they have dimension [nens, nlobs], indexed by (mem_id, obs_id)
    """
    def __init__(self, c, state):
        self.analysis_dir = c.analysis_dir(c.time, c.iter)

        self.info = bcast_by_root(c.comm)(self.parse_obs_info)(c, state)

        self.obs_rec_list = bcast_by_root(c.comm)(self.distribute_obs_tasks)(c)

        self.obs_inds = {}         ##will be created by assimilator.assign_obs()
        self.obs_seq = {}          ##will be created by prepare_obs()
        self.obs_prior_seq = {}    ##will be created by prepare_obs_from_state()
        self.lobs = {}             ##will be created by transpose_to_ensemble_complete()
        self.lobs_prior = {}
        self.lobs_post = {}        ##will be created by assimilator.assimilate()
        self.obs_post_seq = {}     ##will be created by transpose_to_field_complete()
        self.data = {}             ##will be created by pack_obs_data, for use in assimilate()

    def parse_obs_info(self, c, state):
        """
        Parse info for the observation records defined in config.

        Args:
            c (Config): Configuration object.
            state (State): State object.

        Returns:
            dict: A dictionary with some dimensions and list of unique obs records
        """
        info = {'size':0, 'records':{}}
        obs_variables = set()
        obs_err_types = set()
        obs_rec_id = 0  ##record id for an obs sequence
        pos = 0         ##seek position for rec

        ##loop through obs variables defined in obs_def
        for vrec in ensure_list(c.obs_def):
            vname = vrec['name']
            if 'err' not in vrec or vrec['err'] is None:
                vrec['err'] = {}
            assert isinstance(vrec.get('err'), dict), f"obs_def: {vname}: expect 'err' to be a dictionary"
            obs_err_type = vrec['err'].get('type', 'normal')

            ##some properties of the variable is defined in its source module
            dataset = c.datasets[vrec['dataset_src']]
            variables = dataset.variables
            assert vname in variables, 'variable '+vname+' not defined in '+vrec['dataset_src']+'.dataset.variables'

            ##parse impact of obs on each state variable, default is 1.0 on all variables unless set by obs_def record
            impact_on_state = {}
            for state_name in state.info['variables']:
                impact_on_state[state_name] = 1.0
            if 'impact_on_state' in vrec and vrec['impact_on_state'] is not None:
                for state_name, impact_fac in vrec['impact_on_state'].items():
                    impact_on_state[state_name] = impact_fac

            ##loop through time steps in obs window
            for time in c.time + np.array(c.obs_time_steps)*dt1h:
                obs_rec = {'name': vname,
                        'dataset_src': vrec['dataset_src'],
                        'model_src': vrec['model_src'],
                        'nobs': vrec.get('nobs', 0),
                        'obs_window_min': vrec.get('obs_window_min', dataset.obs_window_min),
                        'obs_window_max': vrec.get('obs_window_max', dataset.obs_window_max),
                        'dtype': variables[vname]['dtype'],
                        'is_vector': variables[vname]['is_vector'],
                        'units': variables[vname]['units'],
                        'z_units': variables[vname]['z_units'],
                        'time': time,
                        'dt': 0,
                        'pos': pos,
                        'err':{'type': obs_err_type,
                               'std': vrec['err'].get('std', 1.),  ##for synthetic obs perturb, real obs will have std from dataset
                               'hcorr': vrec['err'].get('hcorr',0.),
                               'vcorr': vrec['err'].get('vcorr',0.),
                               'tcorr': vrec['err'].get('tcorr',0.),
                               'cross_corr': vrec['err'].get('cross_corr',{}),
                               },
                        'hroi': vrec['hroi'] * c.localize_scale_fac[c.iter],
                        'vroi': vrec['vroi'],
                        'troi': vrec['troi'],
                        'impact_on_state': impact_on_state,
                        }
                obs_variables.add(vname)
                obs_err_types.add(obs_err_type)
                info['records'][obs_rec_id] = obs_rec

                ##update obs_rec_id
                obs_rec_id += 1

                ##we don't know the size of obs_seq yet
                ##will wait for prepare_obs to update the seek position

        info['variables'] = list(obs_variables)
        info['err_types'] = list(obs_err_types)

        ##go through the obs_rec again to fill in the default err.cross_corr
        for obs_rec_id, obs_rec in info['records'].items():
            assert isinstance(obs_rec['err']['cross_corr'], dict), f"obs_def: {obs_rec['name']} has err.cross_corr defined as {obs_rec['err']['cross_corr']}, expecting a dictionary"
            for vname in info['variables']:
                if vname not in obs_rec['err']['cross_corr']:
                    if vname == obs_rec['name']:
                        obs_rec['err']['cross_corr'][vname] = 1.0
                    else:
                        obs_rec['err']['cross_corr'][vname] = 0.0
                else:
                    assert isinstance(obs_rec['err']['cross_corr'][vname], float), f"obs_def: {obs_rec['name']} has err.cross_corr.{vname} defined as {obs_rec['err']['cross_corr'][vname]}, expecting a float"

        return info

    def distribute_obs_tasks(self, c):
        """
        Distribute obs_rec_id across processors

        Args:
            c (Config): Configuration object.

        Returns:
            dict: Dictionary {pid_rec (int): list[obs_rec_id (int)]}
        """
        obs_rec_list_full = [i for i in self.info['records'].keys()]
        obs_rec_size = np.array([2 if r['is_vector'] else 1 for i,r in self.info['records'].items()])
        obs_rec_list = distribute_tasks(c.comm_rec, obs_rec_list_full, obs_rec_size)

        return obs_rec_list

    def read_mean_z_coords(self, c, state, time):
        """
        Read the ensemble-mean z coords from z_file at obs time

        Inputs:
            c (Config): Configuration object.
            state (State): State object.
            time (datetime): observation time.

        Returns:
            np.ndarray: Z-coordinate fields of shape (nz, c.grid.x.shape) for all unique levels defined in state.info
        """
        ##first, get a list of indices k
        k_list = list(set([r['k'] for i,r in state.info['fields'].items() if r['time']==time]))

        ##get z coords for each level
        z = np.zeros((len(k_list),)+state.info['shape'])
        for k in range(len(k_list)):

            ##the rec_id in z_file corresponding to this level
            ##there can be multiple records (different state variables)
            ##we take the first one
            rec_id = [i for i,r in state.info['fields'].items() if r['time']==time and r['k']==k_list[k]][0]

            ##read the z field with (mem_id=0, rec_id) from z_file
            z_fld = state.read_field(state.z_coords_file, c.grid.mask, 0, rec_id)

            ##assign the coordinates to z(k)
            if state.info['fields'][rec_id]['is_vector']:
                z[k, ...] = z_fld[0, ...]
            else:
                z[k, ...] = z_fld

        return z

    def state_to_obs(self, c, state, flag, **kwargs):
        """
        Compute the corresponding obs value given the state variable(s), namely the "obs_prior"
        This function includes several ways to compute the obs_prior:

        1, If obs_name is one of the variables provided by the model_src module, then
        model_src.read_var shall be able to provide the obs field defined on model native grid.
        Then we convert the obs field to the analysis grid and do vertical interpolation.

        Note that there are options to obtain the obs fields to speed things up:
        
        1.1, If the obs field can be found in the locally stored fields[mem_id, rec_id],
                we can just access it from memory (quickest)
        
        1.2, If the obs field is not in local memory, but still one of the state variables,
                we can get it through read_field() from the state_file
        
        1.3, If the obs is not one of the state variable, we get it through model_src.read_var()

        2, If obs_name is one of the variables provided by obs.obs_operator, we call it to
        obtain the obs seq. Typically the obs_operator performs more complex computation, such
        as path integration, radiative transfer model, etc. (slowest)

        When using synthetic observations in experiments, this function generates the
        unperturbed obs from the true states (nature runs). Setting "member=None" in the inputs
        indicates that we are trying to generate synthetic observations. Since the state variables
        won't include the true state, the only options above are 1.3 or 2 for this case.

        Args:
            c (Config): config object
            state (State): state object
            flag (str): 'prior' or 'posterior'
            **kwargs: Additional parameters
                - member: int, member index; or None if dealing with synthetic obs
                - name: str, obs variable name
                - time: datetime obj, time of the obs window
                - is_vector: bool, if True the obs is a vector measurement
                - dataset_src: str, dataset source module name providing the obs
                - model_src: str, model source module name providing the state
                - x, y, z, t: np.array, coordinates from obs_seq

        Returns:
            np.ndarray: Values corresponding to the obs_seq but from the state identified by kwargs
        """
        mem_id = kwargs['member']
        synthetic = True if mem_id is None else False
        time = kwargs['time']
        obs_name = kwargs['name']
        is_vector = kwargs['is_vector']

        ##obs dataset source module
        dataset = c.datasets[kwargs['dataset_src']]
        ##model source module
        model = c.models[kwargs['model_src']]

        ##option 1  TODO: update README and comments
        ## if dataset module provides an obs_operator, use it to compute obs
        if hasattr(dataset, 'obs_operator') and obs_name in dataset.obs_operator:
            if synthetic:
                path = model.truth_dir
            else:
                path = c.forecast_dir(time, kwargs['model_src'])

            operator = dataset.obs_operator[kwargs['name']]

            ##get the obs seq from operator
            seq = operator(path=path, model=model, grid=c.grid, mask=c.grid.mask, **kwargs)

        ##option 2:
        ## if obs variable is one of the state variable, or can be computed by the model,
        ## then we just need to collect the 3D variable and interpolate in x,y,z
        elif obs_name in model.variables:

            obs_x = np.array(kwargs['x'])
            obs_y = np.array(kwargs['y'])
            obs_z = np.array(kwargs['z'])
            nobs = len(obs_x)
            if is_vector:
                seq = np.full((2, nobs), np.nan)
            else:
                seq = np.full(nobs, np.nan)

            levels = model.variables[obs_name]['levels']
            for k in range(len(levels)):
                if obs_name in [r['name'] for r in ensure_list(c.state_def)] and not synthetic:
                    ##the obs is one of the state variables
                    ##find its corresponding rec_id
                    rec_id = [i for i,r in state.info['fields'].items() if r['name']==obs_name and r['k']==levels[k]][0]

                    ##option 1.1: if the current pid stores this field, just read it
                    ##TODO: can merge the two into state property getters
                    if rec_id in state.rec_list[c.pid_rec] and mem_id in state.mem_list[c.pid_mem]:
                        z = state.z_fields[mem_id, rec_id]
                        if flag == 'prior':
                            fld = state.fields_prior[mem_id, rec_id]
                        elif flag == 'posterior':
                            fld = state.fields_post[mem_id, rec_id]
                        else:
                            raise ValueError

                    else:  ##option 1.2: read field from state binfile
                        z = state.read_field(state.z_coords_file, c.grid.mask, 0, rec_id)
                        if flag == 'prior':
                            fld = state.read_field(state.prior_file, c.grid.mask, mem_id, rec_id)
                        elif flag == 'posterior':
                            fld = state.read_field(state.post_file, c.grid.mask, mem_id, rec_id)
                        else:
                            raise ValueError

                else:  ##option 1.3: get the field from model.read_var
                    if synthetic:
                        path = model.truth_dir
                    else:
                        path = c.forecast_dir(time, kwargs['model_src'])

                    if k == 0:  ##initialize grid obj for conversion
                        model.read_grid(path=path, **kwargs)
                        model.grid.set_destination_grid(c.grid)

                    model_z = model.z_coords(path=path, member=kwargs['member'], time=kwargs['time'], k=levels[k])
                    z_ = model.grid.convert(model_z, method='nearest')
                    z = np.array([z_, z_]) if is_vector else z_

                    model_fld = model.read_var(path=path, name=kwargs['name'], member=kwargs['member'], time=kwargs['time'], k=levels[k])
                    fld = model.grid.convert(model_fld, is_vector=is_vector, method='nearest')

                ##horizontal interp field to obs_x,y, for current layer k
                if is_vector:
                    z = c.grid.interp(z[0, ...], obs_x, obs_y, method='nearest')
                    zc = np.array([z, z])
                    v1 = c.grid.interp(fld[0, ...], obs_x, obs_y, method='nearest')
                    v2 = c.grid.interp(fld[1, ...], obs_x, obs_y, method='nearest')
                    vc = np.array([v1, v2])
                else:
                    zc = c.grid.interp(z, obs_x, obs_y, method='nearest')
                    vc = c.grid.interp(fld, obs_x, obs_y, method='nearest')

                ##vertical interp to obs_z, take ocean depth as example:
                ##    -------------------------------------------
                ##k-1 -  -  -  -  v[k-1], vp  }dzp  prevous layer
                ##    ----------  z[k-1], zp --------------------
                ##k   -  -  -  -  v[k],   vc  }dzc  current layer
                ##    ----------  z[k],   zc --------------------
                ##k+1 -  -  -  -  v[k+1]
                ##
                ##z at current level k denoted as zc, previous level as zp
                ##variables are considered layer averages, so they are at layer centers
                if k == 0:
                    dzc = zc  ##layer thickness for first level

                    ##the first level: constant vc from z=0 to dzc/2
                    inds = (obs_z >= np.minimum(0, 0.5*dzc)) & (obs_z <= np.maximum(0, 0.5*dzc))
                    seq[..., inds] = vc[..., inds]

                if k > 0:
                    dzc = zc - zp  ##layer thickness for level k

                    ##in between levels: linear interp between vp and vc
                    z_vp = zp - 0.5*dzp
                    z_vc = zp + 0.5*dzc
                    inds = (obs_z >= np.minimum(z_vp, z_vc)) & (obs_z <= np.maximum(z_vp, z_vc))
                    ##there can be collapsed layers if z_vc=z_vp
                    zdiff = z_vc - z_vp
                    collapsed = (zdiff == 0)
                    zdiff = np.where(collapsed, 1, zdiff)
                    vi = ((z_vc - obs_z)*vc + (obs_z - z_vp)*vp) / zdiff
                    vi = np.where(collapsed, vp, vi)

                    ##TODO: still got some warnings here
                    #with np.errstate(divide='ignore'):
                    #    vi = np.where(z_vp==z_vc,
                    #                vp,   ##for collapsed layers just use previous value
                    #                ((z_vc - obs_z)*vc + (obs_z - z_vp)*vp)/(z_vc - z_vp) )  ##otherwise linear interp between layer
                    seq[..., inds] = vi[..., inds]

                if k == len(levels)-1:
                    ##the last level: constant vc from z=zc-dzc/2 to zc
                    inds = (obs_z >= np.minimum(zc-0.5*dzc, zc)) & (obs_z <= np.maximum(zc-0.5*dzc, zc))
                    seq[..., inds] = vc[..., inds]

                if k < len(levels)-1:
                    ##make a copy of current layer as 'previous' for next k
                    zp = zc.copy()
                    vp = vc.copy()
                    dzp = dzc.copy()

        else:
            raise ValueError('unable to obtain obs prior for '+obs_name)

        return seq

    def collect_obs_seq(self, c, state):
        """
        Process the obs in parallel, read dataset files and convert to obs_seq
        which contains obs value, coordinates and other info

        Since this is the actual obs (1 copy), only 1 processor needs to do the work

        Argss:
            c (Config): Configuration object.
            state (State): State object.

        Returns:
            dict: observation sequence. Dictionary {obs_rec_id (int): record}
                where each record is a dictionary {key: np.ndarray}, the mandatory keys are
                'obs' the observed values (measurements)
                'x', 'y', 'z', 't' the coordinates for each measurement
                'err_std' the uncertainties for each measurement
                there can be other optional keys provided by read_obs() but we don't use them
        """
        if c.pid == 0:
            print('>>> read observation sequence from datasets', flush=True)

        ##get obs_seq from dataset module, each pid_rec gets its own workload as a subset of obs_rec_list
        obs_seq = {}
        for obs_rec_id in self.obs_rec_list[c.pid_rec]:
            obs_rec = self.info['records'][obs_rec_id]

            ##load the dataset module
            dataset = c.datasets[obs_rec['dataset_src']]
            assert obs_rec['name'] in dataset.variables, 'variable '+obs_rec['name']+' not defined in dataset.'+obs_rec['dataset_src']+'.variables'

            model = c.models[obs_rec['model_src']]
            ##read ens-mean z coords from z_file for this obs network
            ##typically model.z_coords can compute the z coords as well, but it is more efficient
            ##to just store the ensemble mean z here
            model.z = self.read_mean_z_coords(c, state, obs_rec['time'])

            if c.use_synthetic_obs:
                ##generate synthetic obs network
                seq = dataset.random_network(model=model, grid=c.grid, mask=c.grid.mask, **obs_rec)

                ##compute obs values
                seq['obs'] = self.state_to_obs(c, state, 'prior', member=None, **obs_rec, **seq)

                ##perturb with obs err
                seq['obs'] += np.random.normal(0, 1, seq['obs'].shape) * obs_rec['err']['std']

            else:
                ##read dataset files and obtain obs sequence
                seq = dataset.read_obs(model=model, grid=c.grid, mask=c.grid.mask, **obs_rec)

            if c.pid_mem == 0:
                print('number of '+obs_rec['name']+' obs from '+obs_rec['dataset_src']+': {}'.format(seq['obs'].shape[-1]), flush=True)

            ##misc. transform here
            for transform_func in c.transform_funcs:
                seq = transform_func.forward_obs(c, obs_rec, seq)

            obs_seq[obs_rec_id] = seq
            obs_rec['nobs'] = seq['obs'].shape[-1]  ##update nobs

        ##output obs sequence
        if c.pid_mem == 0:
            for obs_rec_id, rec in obs_seq.items():
                file = os.path.join(self.analysis_dir, f'obs_seq.rec{obs_rec_id}.npy')
                np.save(file, rec)

        return obs_seq

    def prepare_obs(self, c, state):
        self.obs_seq = bcast_by_root(c.comm_mem)(self.collect_obs_seq)(c, state)

    def prepare_obs_from_state(self, c, state, flag):
        """
        Compute the obs priors in parallel, run state_to_obs to obtain obs_prior_seq

        Args:
            c (Config): config object
            state (State): state object
            flag (str): 'prior' or 'posterior'

        Returns:
            dict: obs_prior_seq, Dictionary {(mem_id, obs_rec_id): seq}
                where seq is np.array with values corresponding to obs_seq['obs']
        """
        pid_mem_show = [p for p,lst in state.mem_list.items() if len(lst)>0][0]
        pid_rec_show = [p for p,lst in self.obs_rec_list.items() if len(lst)>0][0]
        c.pid_show =  pid_rec_show * c.nproc_mem + pid_mem_show
        c.print_1p('>>> compute observation priors\n')

        ##process the obs, each proc gets its own workload as a subset of
        ##all proc goes through their own task list simultaneously
        nr = len(self.obs_rec_list[c.pid_rec])
        nm = len(state.mem_list[c.pid_mem])
        for m, mem_id in enumerate(state.mem_list[c.pid_mem]):
            for r, obs_rec_id in enumerate(self.obs_rec_list[c.pid_rec]):
                ##this is the obs record to process
                obs_rec = self.info['records'][obs_rec_id]

                if c.debug:
                    print(f"PID {c.pid:4}: obs_prior mem{mem_id+1:03} {obs_rec['name']:20}", flush=True)
                else:
                    c.print_1p(progress_bar(m*nr+r, nr*nm))

                seq = {}
                ##need the coordinates for transform later
                for key in ['x', 'y', 'z', 't', 'err_std']:
                    seq[key] = self.obs_seq[obs_rec_id][key]
                ##obtain obs_prior values from model state
                seq['obs'] = self.state_to_obs(c, state, flag, member=mem_id, **obs_rec, **self.obs_seq[obs_rec_id])

                ##misc. transform here
                for transform_func in c.transform_funcs:
                    seq = transform_func.forward_obs(c, obs_rec, seq)

                ##collect obs priors together
                if flag == 'prior':
                    self.obs_prior_seq[mem_id, obs_rec_id] = seq['obs']
                elif flag == 'posterior':
                    self.obs_post_seq[mem_id, obs_rec_id] = seq['obs']
                else:
                    raise ValueError
        c.comm.Barrier()
        c.print_1p(' done.\n')

        ##output obs_prior sequeneces
        for key, seq in self.obs_prior_seq.items():
            mem_id, obs_rec_id = key
            file = os.path.join(self.analysis_dir, f'obs_prior_seq.rec{obs_rec_id}.mem{mem_id:03}.npy')
            np.save(file, seq)

    def global_obs_list(self, c):
        ##form the global list of obs (in serial mode the main loop is over this list)
        n_obs_rec = len(self.info['records'])

        i = {}  ##location in full obs vector on owner pid
        for owner_pid in range(c.nproc_mem):
            i[owner_pid] = 0

        obs_list = []
        for obs_rec_id in range(n_obs_rec):
            obs_rec = self.info['records'][obs_rec_id]
            v_list = [0, 1] if obs_rec['is_vector'] else [None]
            for owner_pid in self.obs_inds[obs_rec_id].keys():
                for _ in self.obs_inds[obs_rec_id][owner_pid]:
                    for v in v_list:
                        obs_list.append((obs_rec_id, v, owner_pid, i[owner_pid]))
                        i[owner_pid] += 1

        if getattr(c, 'shuffle_obs', False):
            np.random.shuffle(obs_list)  ##randomize the order of obs (this is optional)

        return obs_list

    def write_obs_info(self, binfile):
        with open(binfile.replace('.bin','.dat'), 'wt') as f:
            f.write('{} {}\n'.format(self.info['nobs'], self.info['nens']))
            for rec in self.info['obs_seq'].values():
                f.write('{} {} {} {} {} {} {} {} {} {} {} {} {} {}\n'.format(rec['name'], rec['dataset_src'], rec['model_src'], rec['dtype'], int(rec['is_vector']), rec['units'], rec['z_units'], rec['x'], rec['y'], rec['z'], t2h(rec['time']), rec['pos']))

    ##read obs_info from the dat file
    def read_obs_info(self, binfile):
        with open(binfile.replace('.bin','.dat'), 'r') as f:
            lines = f.readlines()

            ss = lines[0].split()
            self.info = {'nobs':int(ss[0]), 'nens':int(ss[1]), 'obs_seq':{}}

            ##following lines of obs records
            obs_id = 0
            for lin in lines[1:]:
                ss = lin.split()
                rec = {'name': ss[0],
                    'dataset_src': ss[1],
                    'model_src': ss[2],
                    'dtype': ss[3],
                    'is_vector': bool(int(ss[4])),
                    'units': ss[5],
                    'z_units':ss[6],
                    'err_type': ss[7],
                    'err': np.float32(ss[8]),
                    'x': np.float32(ss[9]),
                    'y': np.float32(ss[10]),
                    'z': np.float32(ss[11]),
                    'time': h2t(np.float32(ss[12])),
                    'pos': int(ss[13]), }
                self.info['obs_seq'][obs_id] = rec
                obs_id += 1

    def transpose_to_ensemble_complete(self, c, state, input_obs, ensemble=False):
        """
        Transpose obs from field-complete to ensemble-complete

        Step 1, Within comm_mem, send the subset of input_obs with mem_id and par_id
        from the source proc (src_pid) to the destination proc (dst_pid), store the
        result in tmp_obs with all the mem_id (ensemble-complete)

        Step 2, Gather all obs_rec_id within comm_rec, so that each pid_rec will have the
        entire obs record for assimilation

        Args:
            c (Config): config obj
            input_obs (dict): obs_seq from process_all_obs() or obs_prior_seq from process_all_obs_priors()
            ensemble (bool): If True, the input_obs is the obs_prior_seq, dict[(mem_id, obs_rec_id), np.array];
                If False (default), the input_obs is the obs_seq, dict[obs_rec_id, dict[key, np.array]]     

        Returns,
            dict: output_obs
                If ensemble, returns the lobs_prior dict[(mem_id, obs_rec_id), dict[par_id, np.array]]
                If not ensemble, returns the lobs dict[obs_rec_id, dict[par_id, dict[key, np.array]]], key = 'obs','x','y','z','t'...
        """
        pid_mem_show = [p for p,lst in state.mem_list.items() if len(lst)>0][0]
        pid_rec_show = [p for p,lst in self.obs_rec_list.items() if len(lst)>0][0]
        c.pid_show =  pid_rec_show * c.nproc_mem + pid_mem_show

        if ensemble:
            c.print_1p('obs prior sequences: ')
        else:
            c.print_1p('obs sequences: ')
        c.print_1p('transpose obs to local obs\n')

        ##Step 1: transpose to ensemble-complete by exchanging mem_id, par_id in comm_mem
        ##        input_obs -> tmp_obs
        tmp_obs = {}  ##local obs at intermediate stage

        nr = len(self.obs_rec_list[c.pid_rec])
        for r, obs_rec_id in enumerate(self.obs_rec_list[c.pid_rec]):

            ##all pid goes through their own mem_list simultaneously
            nm_max = np.max([len(lst) for p,lst in state.mem_list.items()])
            for m in range(nm_max):
                if c.debug:
                    if m < len(state.mem_list[c.pid_mem]):
                        mem_id = state.mem_list[c.pid_mem][m]
                        print(f"PID {c.pid:4}: transposing obs: mem{mem_id+1:03} obs_rec{obs_rec_id}")
                    else:
                        print(f"PID {c.pid:4}: transposing obs: waiting")
                else:
                    c.print_1p(progress_bar(r*nm_max+m, nr*nm_max))

                ##prepare the obs seq for sending if not at the end of mem_list
                if m < len(state.mem_list[c.pid_mem]):
                    mem_id = state.mem_list[c.pid_mem][m]
                    if ensemble:  ##this is the obs prior seq
                        seq = input_obs[mem_id, obs_rec_id].copy()
                    else:
                        if mem_id == 0:  ##this is the obs seq, just let mem_id=0 send it
                            seq = input_obs[obs_rec_id].copy()

                ##the collective send/recv follows the same idea under state.transpose_field_to_state
                ##1) receive lobs_seq from src_pid, for src_pid<pid first
                for src_pid in np.arange(0, c.pid_mem):
                    if m < len(state.mem_list[src_pid]):
                        src_mem_id = state.mem_list[src_pid][m]
                        if ensemble:
                            tmp_obs[src_mem_id, obs_rec_id] = c.comm_mem.recv(source=src_pid, tag=m)
                        else:
                            if src_mem_id == 0:
                                tmp_obs[obs_rec_id] = c.comm_mem.recv(source=src_pid, tag=m)

                ##2) send my obs chunk to a list of dst_pid, send to dst_pid>=pid first
                ##   then cycle back to send to dst_pid<pid. i.e. the dst_pid sequence is
                ##   [pid, pid+1, ..., nproc-1, 0, 1, ..., pid-1]
                if m < len(state.mem_list[c.pid_mem]):
                    for dst_pid in np.mod(np.arange(c.nproc_mem)+c.pid_mem, c.nproc_mem):
                        if ensemble:
                            ##this is the obs prior seq for mem_id, obs_rec_id
                            ##for each par_id, assemble the subset lobs_seq using obs_inds
                            lobs_seq = {}
                            for par_id in state.par_list[dst_pid]:
                                inds = self.obs_inds[obs_rec_id][par_id]
                                lobs_seq[par_id] = seq[..., inds]

                            if dst_pid == c.pid_mem:
                                ##pid already stores the lobs_seq, just copy
                                tmp_obs[mem_id, obs_rec_id] = lobs_seq
                            else:
                                ##send lobs_seq to dst_pid
                                c.comm_mem.send(lobs_seq, dest=dst_pid, tag=m)

                        else:
                            if mem_id == 0:
                                ##this is the obs seq with keys 'obs','err_std','x','y','z','t'
                                ##assemble the lobs_seq dict with same keys but subset obs_inds
                                ##do this for each par_id to get the full lobs_seq
                                lobs_seq = {}
                                for par_id in state.par_list[dst_pid]:
                                    lobs_seq[par_id] = {}
                                    inds = self.obs_inds[obs_rec_id][par_id]
                                    for key in ('obs', 'err_std', 'x', 'y', 'z', 't'):
                                        lobs_seq[par_id][key] = seq[key][..., inds]

                                if dst_pid == c.pid_mem:
                                    ##pid already stores the lobs_seq, just copy
                                    tmp_obs[obs_rec_id] = lobs_seq
                                else:
                                    ##send lobs_seq to dst_pid's lobs
                                    c.comm_mem.send(lobs_seq, dest=dst_pid, tag=m)

                ##3) finish receiving lobs_seq from src_pid, for src_pid>pid now
                for src_pid in np.arange(c.pid_mem+1, c.nproc_mem):
                    if m < len(state.mem_list[src_pid]):
                        src_mem_id = state.mem_list[src_pid][m]
                        if ensemble:
                            tmp_obs[src_mem_id, obs_rec_id] = c.comm_mem.recv(source=src_pid, tag=m)
                        else:
                            if src_mem_id == 0:
                                tmp_obs[obs_rec_id] = c.comm_mem.recv(source=src_pid, tag=m)
        c.comm.Barrier()
        c.print_1p(' done.\n')

        ##Step 2: collect all obs records (all obs_rec_ids) on pid_rec
        ##        tmp_obs -> output_obs
        output_obs = {}
        for entry in c.comm_rec.allgather(tmp_obs):
            for key, data in entry.items():
                output_obs[key] = data
        c.comm.Barrier()

        return output_obs

    def transpose_to_field_complete(self, c, state, lobs):
        """
        Transpose obs from ensemble-complete to field-complete

        Args:
            c (Config): config obj
            state (State): state obj
            lobs (dict): local obs seq

        Returns:
            dict: obs_seq, dictionary {(mem_id, obs_rec_id): np.ndarray}
        """
        pid_mem_show = [p for p,lst in state.mem_list.items() if len(lst)>0][0]
        pid_rec_show = [p for p,lst in self.obs_rec_list.items() if len(lst)>0][0]
        c.pid_show =  pid_rec_show * c.nproc_mem + pid_mem_show

        c.print_1p('obs post sequences: ')
        c.print_1p('transpose local obs to obs\n')

        obs_seq = {}
        nr = len(self.obs_rec_list[c.pid_rec])
        for r, obs_rec_id in enumerate(self.obs_rec_list[c.pid_rec]):

            ##all pid goes through their own mem_list simultaneously
            nm_max = np.max([len(lst) for p,lst in state.mem_list.items()])
            for m in range(nm_max):
                if c.debug:
                    if m < len(state.mem_list[c.pid_mem]):
                        mem_id = state.mem_list[c.pid_mem][m]
                        print(f"PID {c.pid:4}: transposing obs: mem{mem_id+1:03} obs_rec{obs_rec_id}")
                    else:
                        print(f"PID {c.pid:4}: transposing obs: waiting")
                else:
                    c.print_1p(progress_bar(r*nm_max+m, nr*nm_max))

                ##prepare an empty obs_seq for receiving if not at the end of mem_list
                if m < len(state.mem_list[c.pid_mem]):
                    mem_id = state.mem_list[c.pid_mem][m]
                    rec = self.info['records'][obs_rec_id]
                    if rec['is_vector']:
                        seq = np.full((2, rec['nobs']), np.nan)
                    else:
                        seq = np.full((rec['nobs'],), np.nan)

                ##this is just the reverse of transpose_obs_to_lobs
                ## we take the exact steps, but swap send and recv operations here
                ##
                ## 1) send my lobs to dst_pid, for dst_pid<pid first
                for dst_pid in np.arange(0, c.pid_mem):
                    if m < len(state.mem_list[dst_pid]):
                        dst_mem_id = state.mem_list[dst_pid][m]
                        c.comm_mem.send(lobs[dst_mem_id, obs_rec_id], dest=dst_pid, tag=m)

                ## 2) receive fld_chk from a list of src_pid, from src_pid>=pid first
                ##    because they wait to send stuff before able to receive themselves,
                ##    cycle back to receive from src_pid<pid then.
                if m < len(state.mem_list[c.pid_mem]):
                    for src_pid in np.mod(np.arange(c.nproc_mem)+c.pid_mem, c.nproc_mem):

                        if src_pid == c.pid_mem:
                            ##pid already stores the lobs_seq, just copy
                            lobs_seq = lobs[mem_id, obs_rec_id].copy()
                        else:
                            ##send lobs_seq to dst_pid
                            lobs_seq = c.comm_mem.recv(source=src_pid, tag=m)

                        ##unpack the lobs_seq to form a complete seq
                        for par_id in state.par_list[src_pid]:
                            inds = self.obs_inds[obs_rec_id][par_id]
                            seq[..., inds] = lobs_seq[par_id]

                        obs_seq[mem_id, obs_rec_id] = seq

                ## 3) finish sending lobs_seq to dst_pid, for dst_pid>pid now
                for dst_pid in np.arange(c.pid_mem+1, c.nproc_mem):
                    if m < len(state.mem_list[dst_pid]):
                        dst_mem_id = state.mem_list[dst_pid][m]
                        c.comm_mem.send(lobs[dst_mem_id, obs_rec_id], dest=dst_pid, tag=m)
        c.comm.Barrier()
        c.print_1p(' done.\n')
        return obs_seq

    def pack_local_obs_data(self, c, state, par_id, lobs, lobs_prior):
        """pack lobs and lobs_prior into arrays for the jitted functions"""
        data = {}

        ##number of local obs on partition
        nlobs = np.sum([lobs[r][par_id]['obs'].size for r in self.info['records'].keys()])
        n_obs_rec = len(self.info['records'])        ##number of obs records
        n_state_var = len(state.info['variables'])   ##number of state variable names

        data['obs_rec_id'] = np.zeros(nlobs, dtype=int)
        data['obs'] = np.full(nlobs, np.nan)
        data['x'] = np.full(nlobs, np.nan)
        data['y'] = np.full(nlobs, np.nan)
        data['z'] = np.full(nlobs, np.nan)
        data['t'] = np.full(nlobs, np.nan)
        data['err_std'] = np.full(nlobs, np.nan)
        data['obs_prior'] = np.full((c.nens, nlobs), np.nan)
        data['used'] = np.full(nlobs, False)
        data['hroi'] = np.ones(n_obs_rec)
        data['vroi'] = np.ones(n_obs_rec)
        data['troi'] = np.ones(n_obs_rec)
        data['impact_on_state'] = np.ones((n_obs_rec, n_state_var))

        i = 0
        for obs_rec_id in range(n_obs_rec):
            obs_rec = self.info['records'][obs_rec_id]

            data['hroi'][obs_rec_id] = obs_rec['hroi']
            data['vroi'][obs_rec_id] = obs_rec['vroi']
            data['troi'][obs_rec_id] = obs_rec['troi']
            for state_var_id in range(len(state.info['variables'])):
                state_vname = state.info['variables'][state_var_id]
                data['impact_on_state'][obs_rec_id, state_var_id] = obs_rec['impact_on_state'][state_vname]

            local_inds = self.obs_inds[obs_rec_id][par_id]
            d = len(local_inds)
            v_list = [0, 1] if obs_rec['is_vector'] else [None]
            for v in v_list:
                data['obs_rec_id'][i:i+d] = obs_rec_id
                data['obs'][i:i+d] = np.squeeze(lobs[obs_rec_id][par_id]['obs'][v, :])
                data['x'][i:i+d] = lobs[obs_rec_id][par_id]['x']
                data['y'][i:i+d] = lobs[obs_rec_id][par_id]['y']
                data['z'][i:i+d] = lobs[obs_rec_id][par_id]['z'].astype(np.float32)
                data['t'][i:i+d] = np.array([t2h(t) for t in lobs[obs_rec_id][par_id]['t']])
                data['err_std'][i:i+d] = lobs[obs_rec_id][par_id]['err_std']
                for m in range(c.nens):
                    data['obs_prior'][m, i:i+d] = np.squeeze(lobs_prior[m, obs_rec_id][par_id][v, :].copy())
                i += d

        return data

    def unpack_local_obs_data(self, c, state, par_id, lobs, lobs_prior, data):
        """unpack data and write back to the original lobs_prior dict"""
        n_obs_rec = len(self.info['records'])
        i = 0
        for obs_rec_id in range(n_obs_rec):
            obs_rec = self.info['records'][obs_rec_id]

            local_inds = self.obs_inds[obs_rec_id][par_id]
            d = len(local_inds)
            v_list = [0, 1] if obs_rec['is_vector'] else [None]
            for v in v_list:
                for m in range(c.nens):
                    lobs_prior[m, obs_rec_id][par_id][v, :] = data['obs_prior'][m, i:i+d]
                i += d



"""Convert binary state file contents to netcdf files in analysis directory"""
import os
import numpy as np
from NEDAS.utils.netcdf_lib import nc_write_var
from NEDAS.utils.conversion import t2h
from NEDAS.assim_tools.state import read_field

def run(c):
    adir = c.analysis_dir(c.time)
    filename = os.path.join(adir, 'prior_state.bin')

    ##convert bin file state variables [nfield, ny, nx] * nens
    ##to nc files [nens, nt, nz, ny, nx] * num_var files
    info = c.state_info
    dims = {'member':None, 'time':None, 'level':None, 'y':info['ny'], 'x':info['nx']}
    mask = np.full((info['ny'], info['nx']), False, dtype=bool)

    for v in list(set(rec['name'] for i, rec in info['fields'].items())):
        print('converting '+v)
        outfile = filename.replace('.bin','.'+v+'.nc')

        members = np.arange(c.nens)
        times = np.array(list(set(t2h(rec['time']) for i, rec in info['fields'].items() if rec['name']==v)))
        levels = np.array(list(set(rec['k'] for i, rec in info['fields'].items() if rec['name']==v)))

        for rec_id, rec in [(i,r) for i,r in info['fields'].items() if r['name']==v]:
            for mem_id in range(c.nens):
                ##get the field from bin file
                fld = read_field(filename, info, mask, mem_id, rec_id)
                ##get record number along time,level dimensions
                id_time = [i for i,t in enumerate(times) if t2h(rec['time'])==t][0]
                id_level = [i for i,z in enumerate(levels) if rec['k']==z][0]
                recno = {'member':mem_id, 'time':id_time, 'level':id_level}
                if rec['is_vector']:
                    comp = ('_x', '_y')
                    for i in range(2):
                        nc_write_var(outfile, dims, v+comp[i], fld[i,...], recno)
                else:
                    nc_write_var(outfile, dims, v, fld, recno)

        ##output dimensions
        nc_write_var(outfile, {'member':len(members)}, 'member', members, attr={'standard_name':'member', 'long_name':'ensemble member'})
        nc_write_var(outfile, {'time':len(times)}, 'time', times, attr={'standard_name':'time', 'long_name':'time', 'units':'hours since 1900-01-01 00:00:00', 'calendar':'standard'})
        nc_write_var(outfile, {'level':len(levels)}, 'level', levels, attr={'standard_name':'level', 'long_name':'vertical level'})

    ##TODO: output x,y dimensions



import sys
import os
import argparse
import threading
import numpy as np
from netCDF4 import Dataset
from NEDAS.utils.parallel import distribute_tasks, Comm

def read_chunks(filename, chk_list, var_list):
    ##get the total dimensions for each variable
    var_dims = {}
    chunks = {}
    with Dataset(filename+'_0000', 'r') as f:
        for vname in var_list:
            dims = f[vname].dimensions
            if 'west_east' in dims:
                ni = f.getncattr('WEST-EAST_GRID_DIMENSION') - 1
                dim_i = 'west_east'
            elif 'west_east_stag' in dims:
                ni = f.getncattr('WEST-EAST_GRID_DIMENSION')
                dim_i = 'west_east_stag'
            else:
                raise ValueError("west_east dimension not found in "+vname)

            if 'south_north' in dims:
                nj = f.getncattr('SOUTH-NORTH_GRID_DIMENSION') - 1
                dim_j = 'south_north'
            elif 'south_north_stag' in dims:
                nj = f.getncattr('SOUTH-NORTH_GRID_DIMENSION')
                dim_j = 'south_north_stag'
            else:
                raise ValueError("south_north dimension not found in "+vname)

            if 'bottom_top' in dims:
                nk = f.getncattr('BOTTOM-TOP_GRID_DIMENSION') - 1
                dim_k = 'bottom_top'
            elif 'bottom_top_stag' in dims:
                nk = f.getncattr('BOTTOM-TOP_GRID_DIMENSION')
                dim_k = 'bottom_top_stag'
            else:
                nk = 1
                dim_k = 'surface'
            var_dims[vname] = (ni, nj, nk, dim_i, dim_j, dim_k)
            chunks[vname] = {}

    ##each processor read subset of chunks
    for chk_id in chk_list:
        with Dataset(filename+f'_{chk_id:04d}', 'r') as f:
            for vname in var_list:
                dims = f[vname].dimensions
                if 'west_east' in dims:
                    i1 = f.getncattr('WEST-EAST_PATCH_START_UNSTAG') - 1
                    i2 = f.getncattr('WEST-EAST_PATCH_END_UNSTAG')
                elif 'west_east_stag' in dims:
                    i1 = f.getncattr('WEST-EAST_PATCH_START_STAG') - 1
                    i2 = f.getncattr('WEST-EAST_PATCH_END_STAG')

                if 'south_north' in dims:
                    j1 = f.getncattr('SOUTH-NORTH_PATCH_START_UNSTAG') - 1
                    j2 = f.getncattr('SOUTH-NORTH_PATCH_END_UNSTAG')
                elif 'south_north_stag' in dims:
                    j1 = f.getncattr('SOUTH-NORTH_PATCH_START_STAG') - 1
                    j2 = f.getncattr('SOUTH-NORTH_PATCH_END_STAG')

                if 'bottom_top' in dims:
                    k1 = f.getncattr('BOTTOM-TOP_PATCH_START_UNSTAG') - 1
                    k2 = f.getncattr('BOTTOM-TOP_PATCH_END_UNSTAG')
                elif 'bottom_top_stag' in dims:
                    k1 = f.getncattr('BOTTOM-TOP_PATCH_START_STAG') - 1
                    k2 = f.getncattr('BOTTOM-TOP_PATCH_END_STAG')
                else:
                    k1 = 0
                    k2 = 1
                chunks[vname][chk_id] = (i1,i2, j1,j2, k1,k2, f[vname][0, ...])
    return var_dims, chunks

def write_chunks(filename, chk_list, var_list, chunks):
    for chk_id in chk_list:
        with Dataset(filename+f'_{chk_id:04d}', 'r+') as f:
            for vname in var_list:
                _,_, _,_, _,_, chk = chunks[vname][chk_id]
                f[vname][0, ...] = chk

def transpose_chunks_to_fields(comm, chk_list_pid, var_list_pid, var_dims, chunks):
    fields = {}
    pid = comm.Get_rank()
    nproc = comm.Get_size()

    nv_max = np.max([len(lst) for p,lst in var_list_pid.items()])
    for v in range(nv_max):
        if v < len(var_list_pid[pid]):
            ##prepare empty field for receiving chunks
            vname = var_list_pid[pid][v]
            ni, nj, nk, _,_,_ = var_dims[vname]
            fields[vname] = np.full((nk, nj, ni), np.nan)

        for dst_pid in np.arange(0, pid):
            if v < len(var_list_pid[dst_pid]):
                dst_vname = var_list_pid[dst_pid][v]
                comm.send(chunks[dst_vname], dest=dst_pid, tag=v)

        if v < len(var_list_pid[pid]):
            for src_pid in np.mod(np.arange(nproc)+pid, nproc):
                if src_pid == pid:
                    chks = chunks[vname]
                else:
                    chks = comm.recv(source=src_pid, tag=v)

                ##unpack chunks into the full field
                for chk_id in chk_list_pid[src_pid]:
                    i1,i2,j1,j2,k1,k2, chk = chks[chk_id]
                    fields[vname][k1:k2, j1:j2, i1:i2] = chk

        for dst_pid in np.arange(pid+1, nproc):
            if v < len(var_list_pid[dst_pid]):
                dst_vname = var_list_pid[dst_pid][v]
                comm.send(chunks[dst_vname], dest=dst_pid, tag=v)

    return fields

def transpose_fields_to_chunks():

    pass

def read_fields_bin(comm, filename, var_list, var_dims):

    return fields

def write_fields_bin(comm, filename, var_list, var_dims, fields):
    pid = comm.Get_rank()
    if pid == 0:
        with open(filename, 'w') as f:
            pass
    comm.Barrier()
    for vname in var_list:
        with open(filename, 'r+b') as f:
            f.seek()
            f.write()

def read_fields_nc(comm, filename, var_list, var_dims):

    return fields

def write_fields_nc(comm, filename, var_list, var_dims, fields):
    """output a joined nc file for wrfrst, one variable per file"""
    for vname in var_list:
        with Dataset(filename+'_'+vname+'.nc', 'w') as f:
            ni,nj,nk, dim_i,dim_j,dim_k = var_dims[vname]
            if dim_i not in f.dimensions:
                f.createDimension(dim_i, ni)
            if dim_j not in f.dimensions:
                f.createDimension(dim_j, nj)
            if dim_k not in f.dimensions:
                f.createDimension(dim_k, nk)
            f.createVariable(vname, float, (dim_k, dim_j, dim_i))
            f[vname][...] = fields[vname]

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Program to read chunks of wrf restart files into a binary file, or reverse')
    parser.add_argument('mode', choices=['join', 'split'], help='operation mode: join or split')
    parser.add_argument('filename', help='wrf restart file name: wrfrst_d01_<time>')
    parser.add_argument('nchk', default=1, type=int, help='number of processors (chunks)')
    args = parser.parse_args()

    comm = Comm()
    pid = comm.Get_rank()

    chk_list = np.arange(args.nchk)
    chk_list_pid = distribute_tasks(comm, chk_list)

    var_list = ['U_1', 'V_1', 'W_1', 'T', 'PH_1', 'P', 'MU_1', 'QVAPOR', 'QCLOUD']
    var_list_pid = distribute_tasks(comm, var_list)
    full_file = os.path.join(os.path.dirname(args.filename), 'restart')

    if args.mode == 'join':
        var_dims, chunks = read_chunks(args.filename, chk_list_pid[pid], var_list)
        fields = transpose_chunks_to_fields(comm, chk_list_pid, var_list_pid, var_dims, chunks)
        write_fields_nc(comm, full_file, var_list_pid[pid], var_dims, fields)

    # elif args.mode == 'split':
    #     split_chunks(args.filename, chk_list_pid[pid], var_list, full_file)




# Configuration file for the Sphinx documentation builder.
import os
import sys
sys.path.insert(0, os.path.abspath('../..'))

# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'NEDAS'
copyright = '2025, Yue Ying'
author = 'Yue Ying'
release = '1.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx_rtd_theme",
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]

templates_path = ['_templates']
exclude_patterns = []

autodoc_member_order = 'bysource'

autodoc_mock_imports = ["tensorflow", "torch", "opencv", "pygrib", "pyfftw", "numba", "mpi4py"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_show_sphinx = False
html_logo = '../imgs/nedas_logo.png'
html_theme_options = {
    'logo_only': True,
    'includehidden': False,
}

html_static_path = ['_static']



import importlib
from NEDAS.config import Config
from NEDAS.utils.conversion import ensure_list
from .base import Transform

# make sure keys are lower-case
registry = {
    'identity': 'Identity',
    'scale_bandpass': 'ScaleBandpass',
}

def get_transform_funcs(c: Config) -> list[Transform]:
    if c.transform_def is None:
        c.transform_def = {'type':'identity'}
    
    transform_funcs = []
    for transform_func_def in ensure_list(c.transform_def):

        if 'type' not in transform_func_def.keys():
            raise KeyError("'type' needs to be specified in transform_def entries")
        transform_func_type = transform_func_def['type'].lower()

        if transform_func_type not in registry.keys():
            raise NotImplementedError("Transform function type '{transform_func_type}' is not implemented.")
        
        module = importlib.import_module('NEDAS.assim_tools.transforms.'+transform_func_type)
        TransformClass = getattr(module, registry[transform_func_type])
        transform_funcs.append(TransformClass(**transform_func_def))

    return transform_funcs


import numpy as np
from NEDAS.models.vort2d import Vort2DModel
from NEDAS.datasets import Dataset

class Vort2DObs(Dataset):
    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        self.variables = {
            'velocity': {'dtype':'float', 'is_vector':True, 'z_units':'m', 'units':'m/s'},
            'vortex_position': {'dtype':'float', 'is_vector':True, 'z_units':'m', 'units':'m'},
            'vortex_intensity': {'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':'m/s'},
            'vortex_size':  {'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':'m'},
            }

        self.obs_operator = {
            'vortex_position': self.get_vortex_position,
            'vortex_intensity': self.get_vortex_intensity,
            'vortex_size': self.get_vortex_size,
            }

    def random_network(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        name = kwargs['name']
        model = kwargs['model']
        assert isinstance(model, Vort2DModel), 'random_network: ERROR: model must be an instance of Vort2DModel'
        grid = kwargs['grid']

        ##get truth vortex position, some network is vortex-following
        velocity = self.get_velocity(**{**kwargs, 'path': model.truth_dir})
        ##diagnose the vortex position on grid
        i, j = self.vortex_position(velocity[0,...], velocity[1,...])
        true_center_x, true_center_y = grid.x[j,i], grid.y[j,i]

        if 'network_type' in kwargs:
            network_type = kwargs['network_type']
        else:
            network_type = 'global'

        if name == 'velocity':

            nobs = kwargs['nobs']
            if network_type == 'global':
                if nobs is None:
                    nobs = 1000
                y = np.random.uniform(grid.ymin, grid.ymax, nobs)
                x = np.random.uniform(grid.xmin, grid.xmax, nobs)

            elif network_type == 'targeted':
                if nobs is None:
                    nobs = 800   ##note: number of obs in entire domain
                                 ##later only obs within range will be kept
                obs_range = 180000  ##observed range from vortex center, m
                y = np.random.uniform(grid.ymin, grid.ymax, nobs)
                x = np.random.uniform(grid.xmin, grid.xmax, nobs)

                dist = np.hypot(x - true_center_x, y - true_center_y)
                ind = np.where(dist <= obs_range)
                x = x[ind]
                y = y[ind]
                nobs = x.size

            else:
                raise ValueError('unknown network type: '+network_type)

            obs_seq = {'obs': np.full(nobs, np.nan),
                    't': np.full(nobs, kwargs['time']),
                    'z': np.zeros(nobs),
                    'y': y,
                    'x': x,
                    'err_std': np.ones(nobs) * kwargs['err']['std']
                    }

        elif name == 'vortex_position':
            obs_seq = {'obs': np.array([[np.nan, np.nan]]),
                    't': np.array([kwargs['time']]),
                    'z': np.array([0]),
                    'y': np.array([true_center_y]),
                    'x': np.array([true_center_x]),
                    'err_std': np.array([kwargs['err']['std']])
                    }

        elif name in ['vortex_intensity', 'vortex_size']:
            obs_seq = {'obs': np.array([np.nan]),
                    't': np.array([kwargs['time']]),
                    'z': np.array([0]),
                    'y': np.array([true_center_y]),
                    'x': np.array([true_center_x]),
                    'err_std': np.array([kwargs['err']['std']])
                    }

        else:
            raise ValueError('unknown obs variable: '+name)

        return obs_seq

    ###utility functions for obs diagnostics
    def vortex_position(self, u, v):
        ny, nx = u.shape

        ##compute vorticity
        zeta = (np.roll(v, -1, axis=1) - np.roll(v, 1, axis=1) - np.roll(u, -1, axis=0) + np.roll(u, 1, axis=0)) / 2.0

        ##search for max vorticity
        zmax = -999
        center_x, center_y = -1, -1
        buff = 6
        for j in range(buff, ny-buff):
            for i in range(buff, nx-buff):
                z = np.sum(zeta[j-buff:j+buff, i-buff:i+buff])
                if z > zmax:
                    zmax = z
                    center_i, center_j = i, j

        return center_i, center_j

    def vortex_intensity(self, u, v):
        return np.max(np.hypot(u, v))

    def vortex_size(self, u, v, center_i, center_j):
        wind = np.hypot(u, v)
        ny, nx = wind.shape

        nr = 30
        wind_min = 15
        wind_rad = np.zeros(nr)
        count_rad = np.zeros(nr)
        for j in range(-nr, nr+1):
            for i in range(-nr, nr+1):
                r = int(np.sqrt(i**2+j**2))
                if r < nr:
                    wind_rad[r] += wind[int(center_j+j)%ny, int(center_i+i)%nx]
                    count_rad[r] += 1
        wind_rad = wind_rad/count_rad

        if np.max(wind_rad)<wind_min or np.where(wind_rad>=wind_min)[0].size==0:
            Rsize = -1
        else:
            i1 = np.where(wind_rad>=wind_min)[0][-1] ###last point with wind > 35knot
            if i1==nr-1:
                Rsize = i1
            else:
                Rsize = i1 + (wind_rad[i1] - wind_min) / (wind_rad[i1] - wind_rad[i1+1])

        return Rsize

    def get_velocity(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        model = kwargs['model']
        assert isinstance(model, Vort2DModel), 'get_velocity: ERROR: model must be an instance of Vort2DModel'
        grid = kwargs['grid']
        ##read the velocity field from truth
        model_velocity = model.read_var(**{**kwargs, 'name':'velocity'})
        ##convert velocity to target grid
        model.grid.set_destination_grid(grid)
        velocity = model.grid.convert(model_velocity, is_vector=True)
        return velocity

    def get_vortex_position(self, **kwargs):
        velocity = self.get_velocity(**kwargs)
        grid = kwargs['model'].grid
        center_i, center_j = self.vortex_position(velocity[0,...], velocity[1,...])
        obs_seq = np.zeros((2, 1), dtype='float')
        obs_seq[0,0] = grid.x[center_j, center_i]
        obs_seq[1,0] = grid.y[center_j, center_i]
        return obs_seq

    def get_vortex_intensity(self, **kwargs):
        velocity = self.get_velocity(**kwargs)
        Vmax = self.vortex_intensity(velocity[0,...], velocity[1,...])
        return np.array([Vmax])

    def get_vortex_size(self, **kwargs):
        velocity = self.get_velocity(**kwargs)
        dx = kwargs['model'].grid.dx
        center_i, center_j = self.vortex_position(velocity[0,...], velocity[1,...])
        Rsize = self.vortex_size(velocity[0,...], velocity[1,...], center_i, center_j)
        Rsize = Rsize * dx
        return np.array([Rsize])

    def read_obs(self):
        raise NotImplementedError("read_obs is not implemented for vort2d, since only using synthetic obs.")


import copy
from abc import abstractmethod
import numpy as np
from NEDAS.utils.parallel import bcast_by_root, distribute_tasks
from NEDAS.utils.progress import progress_bar
from NEDAS.assim_tools.assimilators.base import Assimilator

class SerialAssimilator(Assimilator):
    """
    Subclass for serial assimilation algorithms
    """

    def init_partitions(self, c):
        """
        Generate spatial partitioning of the domain
        """
        if len(c.grid.x.shape) == 2:
            ny, nx = c.grid.x.shape
            ##the domain is divided into tiles, each is formed by nproc_mem elements
            ##each element is stored on a different pid_mem
            ##for each pid, its loc points cover the entire domain with some spacing

            ##list of possible factoring of nproc_mem = nx_intv * ny_intv
            ##pick the last factoring that is most 'square', so that the interval
            ##is relatively even in both directions for each pid
            nx_intv, ny_intv = [(i, int(c.nproc_mem / i))
                                for i in range(1, int(np.ceil(np.sqrt(c.nproc_mem))) + 1)
                                if c.nproc_mem % i == 0][-1]

            ##a list of (ist, ied, di, jst, jed, dj) for slicing
            ##note: we have nproc_mem entries in the list
            partitions = [(i, nx, nx_intv, j, ny, ny_intv)
                        for j in range(ny_intv) for i in range(nx_intv) ]

        else:
            npoints = c.grid.x.size
            ##just divide the list of points into nproc_mem parts, each part spanning the entire domain
            nparts = c.nproc_mem
            partitions = [np.arange(i, npoints, nparts) for i in np.arange(nparts)]

        return partitions

    def assign_obs(self, c, state, obs):
        """
        Assign the observation sequence to each partition par_id

        Args:
            c (Config): configuration object
            state (State): state object
            obs (Obs): observation object

        Returns:
            dict[obs_rec_id, dict[par_id, np.ndarray]]:
               Indices in the full obs_seq for the subset of obs that belongs to partition par_id
        """
        obs_inds_pid = {}
        for obs_rec_id in obs.obs_rec_list[c.pid_rec]:
            full_inds = np.arange(obs.obs_seq[obs_rec_id]['obs'].shape[-1])
            obs_inds_pid[obs_rec_id] = {}

            ##locality doesn't matter, we just divide obs_rec into nproc_mem parts
            inds = distribute_tasks(c.comm_mem, full_inds)
            for par_id in range(c.nproc_mem):
                obs_inds_pid[obs_rec_id][par_id] = inds[par_id]

        ##now each pid_rec has figured out obs_inds for its own list of obs_rec_ids, we
        ##gather all obs_rec_id from different pid_rec to form the complete obs_inds dict
        obs_inds = {}
        for entry in c.comm_rec.allgather(obs_inds_pid):
            for obs_rec_id, data in entry.items():
                obs_inds[obs_rec_id] = data

        return obs_inds

    def distribute_partitions(self, c, state, obs):
        ##just assign each partition to each pid, pid==par_id
        par_list = {p:np.array([p]) for p in range(c.nproc_mem)}
        return par_list

    def assimilation_algorithm(self, c, state, obs):
        """
        Implementation of the serial assimilation algorithm.

        Notes:
            serial assimilation goes through the list of observations one by one
            for each obs the near by state variables are updated one by one.
            so each update is a scalar problem, which is solved in 2 steps: obs_increment, update_ensemble
        """
        state.state_post = copy.deepcopy(state.state_prior)
        obs.lobs_post =copy.deepcopy(obs.lobs_prior)

        par_id = c.pid_mem

        state_data = state.pack_local_state_data(c, par_id, state.state_prior, state.z_state)

        obs_data = obs.pack_local_obs_data(c, state, par_id, obs.lobs, obs.lobs_prior)
        obs_list = bcast_by_root(c.comm)(obs.global_obs_list)(c)

        c.print_1p('>>> assimilate in serial mode:\n')
        ##go through the entire obs list, indexed by p, one scalar obs at a time
        for p in range(len(obs_list)):
            obs_rec_id, v, owner_pid, i = obs_list[p]

            if c.debug:
                print(f"Processing observation obs_rec_id={obs_rec_id:2}, i={i}", flush=True)
            else:
                c.print_1p(progress_bar(p, len(obs_list)))

            ##1. if the pid owns this obs, broadcast it to all pid
            if c.pid_mem == owner_pid:
                ##collect obs info
                obs_p = {}
                obs_p['prior'] = obs_data['obs_prior'][:, i]
                for key in ('obs', 'x', 'y', 'z', 't', 'err_std'):
                    obs_p[key] = obs_data[key][i]
                for key in ('hroi', 'vroi', 'troi', 'impact_on_state'):
                    obs_p[key] = obs_data[key][obs_rec_id]
                ##mark this obs as used
                obs_data['used'][i] = True

            else:
                obs_p = None
            obs_p = c.comm_mem.bcast(obs_p, root=owner_pid)

            if np.isnan(obs_p['prior']).any() or np.isnan(obs_p['obs']):
                continue

            ##compute obs-space increment
            obs_incr = self.obs_increment(obs_p['prior'], obs_p['obs'], obs_p['err_std'])

            ##2. all pid update their own locally stored state:
            state_h_dist = c.grid.distance(obs_p['x'], state_data['x'], obs_p['y'], state_data['y'], p=2)
            state_v_dist = np.abs(obs_p['z'] - state_data['z'])
            state_t_dist = np.abs(obs_p['t'] - state_data['t'])
            self.update_local_state(state_data['state_prior'], obs_p['prior'], obs_incr,
                            state_h_dist, state_v_dist, state_t_dist,
                            obs_p['hroi'], obs_p['vroi'], obs_p['troi'],
                            c.localization_funcs['horizontal'], c.localization_funcs['vertical'], c.localization_funcs['temporal'])

            ##3. all pid update their own locally stored obs:
            obs_h_dist = c.grid.distance(obs_p['x'], obs_data['x'], obs_p['y'], obs_data['y'], p=2)
            obs_v_dist = np.abs(obs_p['z'] - obs_data['z'])
            obs_t_dist = np.abs(obs_p['t'] - obs_data['t'])
            self.update_local_obs(obs_data['obs_prior'], obs_data['used'], obs_p['prior'], obs_incr,
                            obs_h_dist, obs_v_dist, obs_t_dist,
                            obs_p['hroi'], obs_p['vroi'], obs_p['troi'],
                            c.localization_funcs['horizontal'], c.localization_funcs['vertical'], c.localization_funcs['temporal'])

        state.unpack_local_state_data(c, par_id, state.state_post, state_data)
        obs.unpack_local_obs_data(c, state, par_id, obs.lobs, obs.lobs_post, obs_data)
        c.print_1p(' done.\n')

    @abstractmethod
    def obs_increment(self, obs_prior, obs, obs_err) -> np.ndarray:
        """
        Compute observation-space analysis increments.

        Args:
            obs_prior (np.ndarray): Observation priors, 1D array of type int size nens
            obs (float): The real observation value
            obs_err (float): Observation error std

        Returns:
            ndarray: observation-space analysis increments
        """
        pass

    @abstractmethod
    def update_local_state(self, state_prior, obs_prior, obs_incr,
                           state_h_dist, state_v_dist, state_t_dist,
                           hroi, vroi, troi,
                           h_local_func, v_local_func, t_local_func) -> None:
        """
        Update the local state vector with the analysis increments.

        Args:
            state_data (np.ndarray[(nens, nfld, nloc), float]): Local state vector
            obs_prior (np.ndarray[nens, float]): Observation priors
            obs_incr (np.ndarray[nens, float]): Analysis increments

        """
        pass

    @abstractmethod
    def update_local_obs(self, obs_data_prior, obs_used, obs_prior, obs_incr,
                         obs_h_dist, obs_v_dist, obs_t_dist,
                         hroi, vroi, troi,
                         h_local_func, v_local_func, t_local_func) -> None:
        """
        Update the local observations with analysis increments.

        Args:
            obs_data_prior (np.ndarray):
            obs_used (np.nd
        """
        pass



import os
import subprocess
from NEDAS.job_submitters import get_job_submitter

def makedir(dirname:str) -> None:
    """
    Create a directory if it does not exist.
 
    FileExistsError can happen if multiple processors are trying to make the same directory.
    This function will ignore this error and continue.

    Args:
        dirname (str): Directory name to be created.
    """
    try:
        os.makedirs(dirname, exist_ok=True)
    except FileExistsError:
        pass

def run_command(shell_cmd:str) -> None:
    """
    Run a shell command in a subprocess and check for errors.

    Args:
        shell_cmd (str): Shell command to be executed.

    Raises:
        RuntimeError: If the command returns a non-zero exit status.
    """
    p = subprocess.run(shell_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

    if p.returncode != 0:
        raise RuntimeError(f"{p.stderr}")

def run_job(commands:str, **kwargs):
     """
     Run a shell command by submitting it to a job scheduler using JobSubmitter class.
     
     Args:
        commands (str): Shell command to be executed.
        **kwargs: Key-value pairs to passed to the job submitter run method.
     """
     ##get the right job submitter
     job_submitter = get_job_submitter(**kwargs)

     ##run job using the submitter
     job_submitter.run(commands)



import numpy as np
from NEDAS.datasets import Dataset

class QGObs(Dataset):
    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        self.variables = {
            'velocity': {'dtype':'float', 'is_vector':True, 'z_units':'*', 'units':'*'},
            'streamfunc': {'dtype':'float', 'is_vector':False, 'z_units':'*', 'units':'*'},
            'temperature': {'dtype':'float', 'is_vector':False, 'z_units':'*', 'units':'*'},
            'vorticity': {'dtype':'float', 'is_vector':False, 'z_units':'*', 'units':'*'},
        }

        self.operator = {}

    def random_network(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)

        if kwargs['nobs'] is None:
            nobs = 1000
        else:
            nobs = kwargs['nobs']

        grid = kwargs['grid']

        obs_y = np.random.uniform(grid.ymin, grid.ymax, nobs)
        obs_x = np.random.uniform(grid.xmin, grid.xmax, nobs)

        # obs_z = np.random.uniform(0, 1, self.nobs)
        obs_z = np.zeros(nobs)

        obs_seq = {'obs': np.full(nobs, np.nan),
                't': np.full(nobs, kwargs['time']),
                'z': obs_z,
                'y': obs_y,
                'x': obs_x,
                'err_std': np.ones(nobs) * kwargs['err']['std']
                }
        return obs_seq
    
    def read_obs(self, **kwargs):
        raise NotImplementedError('read_obs: not implemented for qg model since only using synthetic obs')



import os
import subprocess
from datetime import timedelta
from NEDAS.config import Config
from NEDAS.schemes.offline_filter import OfflineFilterAnalysisScheme

def generate_init_ensemble(c: Config) -> None:
    scheme = OfflineFilterAnalysisScheme()
    model = c.models['qg']

    print(f"Creating initial condition for qg model:")
    init_time = c.time
    c.time -= model.spinup_hours * timedelta(hours=1)
    c.cycle_period = model.spinup_hours
    print(f"initial condition type: {model.psi_init_type}")
    print(f"spinup period: {model.spinup_hours} hours")

    print(f"Running ensemble forecast from {c.time} to {c.next_time}")
    scheme.run_step(c, 'ensemble_forecast')

    print("Moving output files")
    fcst_dir = c.forecast_dir(c.time, 'qg')
    cycle_dir = c.cycle_dir(c.time)
    ens_init_dir = model.ens_init_dir
    basename = f"output_{init_time:%Y%m%d_%H}.bin"
    for m in range(c.nens):
        mstr = f"{m+1:04d}"

        src_file = os.path.join(fcst_dir, mstr, basename)
        init_file = os.path.join(ens_init_dir, mstr, basename)

        os.makedirs(os.path.dirname(init_file), exist_ok=True)
        subprocess.run(f"mv -v {src_file} {init_file}", shell=True, check=True)

    print(f"removing temporary run directory: {cycle_dir}")
    os.system(f"rm -rf {cycle_dir}")

if __name__ == '__main__':
    c = Config(parse_args=True)
    generate_init_ensemble()



import os
from NEDAS.config import Config

def generate_init_ensemble(c: Config) -> None:
    model = c.models['vort2d']

    ens_init_dir = model.ens_init_dir
    os.system("mkdir -p "+ens_init_dir)

    for m in range(c.nens):
        opts = {
            'path': ens_init_dir,
            'name': 'velocity',
            'is_vector': True,
            'time': c.time,
            'member': m,
            }
        print(f"generating initial condition for member {m+1}, output to {model.filename(**opts)}")

        state = model.generate_initial_condition()

        model.write_var(state, **opts)

    print("done.")

if __name__ == '__main__':
    c = Config(parse_args=True)
    generate_init_ensemble()



import numpy as np
from matplotlib import cm
from matplotlib import colormaps
from matplotlib.colors import BoundaryNorm
from matplotlib.path import Path
from matplotlib.patches import PathPatch

def get_cmap(cmap_name: str):
    """
    Get colormap object based on the input name string.

    Args:
        cmap_name (str): The name of the color map. For `cmocean` colormaps, the name
            should be in the format 'cmocean.<cmap_name>'

    Returns:
        Colormap: A colormap object corresponding to the given name.

    Raises:
        KeyError: If the colormap name is not found.
    """
    if cmap_name.split('.')[0] == 'cmocean':
        import cmocean
        cmap = getattr(cmocean.cm, cmap_name.split('.')[-1])
    else:
        cmap = colormaps[cmap_name]
    return cmap

def adjust_ax_size(ax):
    """
    Make plot axes a little smaller on right hand side to make room for colorbar.

    Even for axes without colorbar in a multi-pane plot, it is still useful to call this function
    so that the axes will align with those with colorbars.

    Args:
        ax (matplotlib.axes.Axes): Matplotlib axes object.
    """
    left, bottom, width, height = ax.get_position().bounds
    ax.set_position([left, bottom, width*0.9, height])

def add_colorbar(fig, ax, cmap, vmin, vmax, nlevels=10, fontsize=12, units=None):
    """
	Add a colorbar to the right-hand side of an axes.

    This function adds a colorbar to the provided axes, using the specified
    colormap and value range. The number of levels, font size, and unit label
    can be customized.

    Args:
        fig (matplotlib.figure.Figure): Matplotlib figure object.
        ax (matplotlib.axes.Axes): Matplotlib axes object.
        cmap (matplotlib.colors.Colormap): Colormap to use for the colorbar.
        vmin (float): Minimum value for the colorbar.
        vmax (float): Maximum value for the colorbar.
        nlevels (int, optional): Number of color levels. Defaults to 10.
        fontsize (int, optional): Font size for tick labels and unit label. Defaults to 12.
        units (str, optional): Unit label to display as the colorbar title. Defaults to None.

    Returns:
        matplotlib.colorbar.Colorbar: The created colorbar object.
    """
    dv = (vmax - vmin) / nlevels
    bounds = np.arange(vmin, vmax+dv, dv)

    norm = BoundaryNorm(bounds, ncolors=256, extend='both')

    ##adjust the main plot ax to make room for colorbar
    left, bottom, width, height = ax.get_position().bounds
    ax.set_position([left, bottom, width*0.9, height])

    ##add colorbar ax to the right
    cax = fig.add_axes([left+width*0.95, bottom+height*0.15, width*0.03, height*0.7])
    cax.tick_params(labelsize=fontsize)

    ##draw the colorbar
    cbar = fig.colorbar(cm.ScalarMappable(cmap=cmap, norm=norm), cax=cax, ticks=bounds)
    if units is not None:
        cbar.ax.set_title(units, fontsize=fontsize, loc='left', pad=25)
    return cbar

def arrowhead_xy(x1, x2, y1, y2, hw, hl):
    """
	Given a line segment from (x1,y1) to (x2,y2), return the segments that draw an arrow head (for plotting vectors).

    Args:
        x1 (float): X-coordinate of the start point of line segment.
        x2 (float): X-coordinate of the end point of line segment.
        y1 (float): Y-coordinate of the start point of line segment.
        y2 (float): Y-coordinate of the end point of line segment.
        hw (float): Width of the arrow head.
        hl (float): Length of the arrow head.

    Returns:
        list: X-coordinates of the additional line segments forming the arrowhead.
        list: Y-coordinates of the additional line segments forming the arrowhead.
    """
    np.seterr(invalid='ignore')
    ll = np.sqrt((x2 - x1)**2 + (y2 - y1)**2)
    sinA = (y2 - y1)/ll
    cosA = (x2 - x1)/ll
    h1x = x1 - 0.2*hl*cosA
    h1y = y1 - 0.2*hl*sinA
    h2x = x1 + 0.8*hl*cosA - 0.5*hw*sinA
    h2y = y1 + 0.8*hl*sinA + 0.5*hw*cosA
    h3x = x1 + 0.5*hl*cosA
    h3y = y1 + 0.5*hl*sinA
    h4x = x1 + 0.8*hl*cosA + 0.5*hw*sinA
    h4y = y1 + 0.8*hl*sinA - 0.5*hw*cosA
    return [h1x, h2x, h3x, h4x, h1x], [h1y, h2y, h3y, h4y, h1y]

def draw_reference_vector_legend(ax, xr, yr, V, L, hw, hl, refcolor, linecolor, ref_units=''):
    """
    Draw a legend box with reference vector and units string.

    Args:
        ax (matplotlib.axes.Axes): Matplotlib axes object.
        xr (float): X-coordinate of the center of the legend box
        yr (float): Y-coordinate of the center of the legend box
        V (float): Velocity scale to be shown as the reference vector.
        L (float): Length of the reference vector.
        hw (float): Width of the arrowhead.
        hl (float): Length of the arrowhead.
        refcolor (str or tuple): Color of the background in the legend box.
        linecolor (str or tuple): Color of the reference vector.
        ref_units (str, optional): Unit label to be shown, default is ''.
    """
    ##draw a box
    xb = [xr-L*1.3, xr-L*1.3, xr+L*1.3, xr+L*1.3, xr-L*1.3]
    yb = [yr+L/2, yr-L, yr-L, yr+L/2, yr+L/2]
    ax.fill(xb, yb, color=refcolor, zorder=6)
    ax.plot(xb, yb, color='k', zorder=6)
    ##draw the reference vector
    ax.plot([xr-L/2, xr+L/2], [yr, yr], color=linecolor, zorder=8)
    ax.fill(*arrowhead_xy(xr+L/2, xr-L/2, yr, yr, hw, hl), color=linecolor, zorder=8)
    ##add unit string annotation below the vector
    ax.text(xr, yr-L/2, f"{V} {ref_units}", color=linecolor, ha='center', va='center', zorder=8)

def draw_line(ax, data, linecolor, linewidth, linestyle, zorder):
    """
    Draw line segments.

    Args:
        ax (matplotlib.axes.Axes): The axes on which to draw the lines.
        data (dict): Dictionary containing:
            - 'xy' (list of arrays): Coordinates of points.
            - 'parts' (list of lists): Indices indicating segment divisions.
        linecolor (str or tuple): Color of the line.
        linewidth (float): Width of the line.
        linestyle (str): Style of the line (e.g., '-', '--', ':').
        zorder (int): Drawing order (higher numbers are drawn on top).
    """
    xy = data['xy']
    parts = data['parts']
    for i in range(len(xy)):
        for j in range(len(parts[i])-1): ##plot separate segments if multi-parts
            ax.plot(*zip(*xy[i][parts[i][j]:parts[i][j+1]]), color=linecolor, linewidth=linewidth, linestyle=linestyle, zorder=zorder)
        ax.plot(*zip(*xy[i][parts[i][-1]:]), color=linecolor, linewidth=linewidth, linestyle=linestyle, zorder=zorder)

def draw_patch(ax, data, color, zorder):
    """
    Draw a filled patch.

    Args:
        ax (matplotlib.axes.Axes): The axes on which to draw the lines.
        data (dict): Dictionary containing:
            - 'xy' (list of arrays): Coordinates of points.
            - 'parts' (list of lists): Indices indicating segment divisions.
        color (str or tuple): Color of the patch
        zorder (int): Drawing order (higher numbers are drawn on top).
    """
    xy = data['xy']
    parts = data['parts']
    for i in range(len(xy)):
        code = [Path.LINETO] * len(xy[i])
        for j in parts[i]:  ##make discontinuous patch if multi-parts
            code[j] = Path.MOVETO
        ax.add_patch(PathPatch(Path(xy[i], code), facecolor=color, edgecolor=color, linewidth=0.1, zorder=zorder))



import numpy as np
from .base import Updator

class AdditiveUpdator(Updator):

    def compute_increment(self, c, state):
        """
        Additive updator: just compute the difference between prior and posterior as increments
        """
        self.increment = {}
        for mem_id, rec_id in state.fields_prior.keys():
            rec = state.info['fields'][rec_id]
            fld_prior = state.fields_prior[mem_id, rec_id]
            fld_post = state.fields_post[mem_id, rec_id]

            ##misc transform inverse
            for transform_func in c.transform_funcs:
                fld_prior = transform_func.backward_state(c, rec, fld_prior)
                fld_post = transform_func.backward_state(c, rec, fld_post)

            ##collect the increments
            self.increment[mem_id, rec_id] = fld_post - fld_prior

    def update_restartfile(self, c, state, mem_id, rec_id):
        """
        Method to update a single field rec_id in the model restart file.
        This can be overridden by derived classes for specific update methods
        Inputs:
        - c: config object
        - state: state object
        - mem_id: member index
        - rec_id: record index
        """
        rec = state.info['fields'][rec_id]
        model = c.models[rec['model_src']]
        path = c.forecast_dir(rec['time'], rec['model_src'])
        model.read_grid(path=path, member=mem_id, **rec)

        ##convert the posterior variable back to native model grid
        var_prior = model.read_var(path=path, member=mem_id, **rec)

        c.grid.set_destination_grid(model.grid)
        incr = c.grid.convert(self.increment[mem_id, rec_id], is_vector=rec['is_vector'], method='linear')

        if rec['is_vector']:
            fld_shape = var_prior.shape[1:]
        else:
            fld_shape = var_prior.shape

        if fld_shape == model.grid.x.shape:
            var_post = var_prior + incr
        elif fld_shape == model.grid.x_elem.shape:
            incr = np.mean(incr[...,model.grid.tri.triangles], axis=-1)
            var_post = var_prior + incr
        else:
            raise RuntimeError(f"mismatch in field prior {var_prior.shape} with increment {incr.shape}")

        ##TODO: temporary solution for nan values due to interpolation
        ind = np.where(np.isnan(var_post))
        var_post[ind] = var_prior[ind]
        # if np.isnan(var_post).any():
        #     raise ValueError('nan detected in var_post')

        ##write the posterior variable to restart filediff
        model.write_var(var_post, path=path, member=mem_id, comm=c.comm, **rec)



import numpy as np
from NEDAS.grid import Grid
from NEDAS.utils.fft_lib import fft2, ifft2, get_wn

"""functions to perform scale decomposition, through bandpass filtering
on spatial data (fields), can also use gcm_filter to achieve this.
"""
def lowpass_response(k2d, k1, k2):
    """
    Low-pass spectral response function
    Input
    - k2d: array, 2d wavenumber
    - k1, k2: float, wavenumbers defining the transition zone (from 1 to 0 in response)
    Return
    - r: array same dimension as k2d, the response function
    """
    r = np.zeros(k2d.shape)
    r[np.where(k2d<k1)] = 1.0
    r[np.where(k2d>k2)] = 0.0
    ind = np.where(np.logical_and(k2d>=k1, k2d<=k2))
    ##cos-square transition
    r[ind] = np.cos((k2d[ind] - k1)*(0.5*np.pi/(k2 - k1)))**2
    return r

def get_scale_component_spec_bandpass(grid, fld, character_length, s):
    assert grid.regular, "get_scale_component_spec_bandpass only works for regular grid"

    ##convert length to wavenumber
    L = max(grid.Lx, grid.Ly)
    character_k = L / np.array(character_length)

    nscale = len(character_k)
    if nscale == 1:
        return fld ##nothing to be done, return the original field

    kx, ky = get_wn(fld)
    k2d = np.hypot(kx, ky)
    fld_spec = fft2(fld)
    if s == 0:
        r = lowpass_response(k2d, character_k[s], character_k[s+1])
    if s == nscale-1:
        r = 1 - lowpass_response(k2d, character_k[s-1], character_k[s])
    if s > 0 and s < nscale-1:
        r = lowpass_response(k2d, character_k[s], character_k[s+1]) - lowpass_response(k2d, character_k[s-1], character_k[s])

    return ifft2(fld_spec * r)

def convolve(grid, fld, rgrid, response):
    L = max(grid.Lx, grid.Ly)
    kx, ky = get_wn(rgrid.x)
    ny, nx = rgrid.x.shape
    fld1 = fld.copy()
    ##go through grid points in the field and perform convolution
    for i in np.ndindex(grid.x.shape):
        ##shift to the grid point i position
        r = response * np.exp(-2 * np.pi * complex(0,1) * (kx*grid.x[i] + ky*grid.y[i]) / L)
        ##convert to physical space
        kernel = ifft2(r) * nx * ny / grid.x.size
        w = rgrid.convert(kernel)
        ##perform convolution
        fld1[i] = np.nansum(fld * w)
    return fld1

def get_scale_component_convolve(grid, fld, character_length, s):
    ##convert length to wavenumber
    L = max(grid.Lx, grid.Ly)
    character_k = L / np.array(character_length)

    nscale = len(character_k)
    if nscale == 1:
        ##nothing to be done, return the original field
        return fld

    ##make a regular grid for the kernel function
    rgrid = Grid.regular_grid(grid.proj, grid.xmin, grid.xmax, grid.ymin, grid.ymax, grid.dx)
    rgrid.set_destination_grid(grid)
    kx, ky = get_wn(rgrid.x)
    ny, nx = rgrid.x.shape
    k2d = np.hypot(kx, ky)

    if s == 0:
        r = lowpass_response(k2d, character_k[s], character_k[s+1])
        return convolve(grid, fld, rgrid, r)
    if s == nscale-1:
        r = lowpass_response(k2d, character_k[s-1], character_k[s])
        return fld - convolve(grid, fld, rgrid, r)
    if s > 0 and s < nscale-1:
        r1 = lowpass_response(k2d, character_k[s-1], character_k[s])
        r2 = lowpass_response(k2d, character_k[s], character_k[s+1])
        return convolve(grid, fld, rgrid, r2) - convolve(grid, fld, rgrid, r1)

def get_scale_component(grid, fld, character_length, s):
    """
    Get scale component using a bandpass filter in spectral space
    Input:
    - grid: Grid object
    - fld: array, [..., ny, nx], the input field
    - character_length: list of characteristic length for each scale
    - s: int, scale index
    Return:
    - fld: array, the scale component s of input fld
    """
    flds = fld.copy()
    if grid.regular:
        for i in np.ndindex(fld.shape[:-2]):
            flds[i] = get_scale_component_spec_bandpass(grid, fld[i], character_length, s)
    else:
        for i in np.ndindex(fld.shape[:-1]):
            flds[i] = get_scale_component_convolve(grid, fld[i], character_length, s)
    return flds

def get_error_scale_factor(grid, character_length, s):
    err_scale_fac = np.ones(grid.x.shape)
    L = max(grid.Lx, grid.Ly)
    character_k = L / np.array(character_length)
    nscale = len(character_k)
    if nscale == 1:
        return err_scale_fac

    rgrid = Grid.regular_grid(grid.proj, grid.xmin, grid.xmax, grid.ymin, grid.ymax, grid.dx)
    rgrid.set_destination_grid(grid)
    kx, ky = get_wn(rgrid.x)
    ny, nx = rgrid.x.shape
    k2d = np.hypot(kx, ky)
    if s == 0:
        response = lowpass_response(k2d, character_k[s], character_k[s+1])
    if s == nscale-1:
        response = 1 - lowpass_response(k2d, character_k[s-1], character_k[s])
    if s > 0 and s < nscale-1:
        r1 = lowpass_response(k2d, character_k[s-1], character_k[s])
        r2 = lowpass_response(k2d, character_k[s], character_k[s+1])
        response = r2 - r1

    for i in np.ndindex(grid.x.shape):
        ##shift to the grid point i position
        r = response * np.exp(-2 * np.pi * complex(0,1) * (kx*grid.x[i] + ky*grid.y[i]) / L)
        ##convert to physical space
        kernel = ifft2(r) * nx * ny / grid.x.size
        w = rgrid.convert(kernel)
        err_scale_fac[i] = w[i]
    return err_scale_fac




from typing import Optional, Dict, Literal
import numpy as np
from netCDF4 import Dataset
from NEDAS.utils.parallel import Comm

AccessMode = Literal['r', 'w', 'a', 'r+']

def nc_open(filename: str, mode: AccessMode, comm: Optional[Comm]=None) -> Dataset:
    """
    Open a netCDF file.

    Args:
        filename (str): Path to the netCDF file.
        mode (str): Open mode, (e.g. :code:`'r'`, :code:`'w'`).
        comm (Comm, optional): MPI communicator object. If None, open the netCDF4.Dataset normally.
            If communicator is available, try :code:`parallel=True` open when opening the file.
            If that's not supported, acquire a file lock in communicator for blocking serial access of the file.

    Returns:
        netCDF4.Dataset: netCDF file handle.
    """
    if comm is None:
        return Dataset(filename, mode, format='NETCDF4')
    else:
        if comm.parallel_io:
            return Dataset(filename, mode, format='NETCDF4', parallel=True)
        else:
            comm.acquire_file_lock(filename)
            try:
                return Dataset(filename, mode, format='NETCDF4')
            except Exception:
                comm.release_file_lock(filename)
                raise

def nc_close(filename: str, f: Dataset, comm: Optional[Comm]=None) -> None:
    """
    Close the netCDF file handle.

    Args:
        filename (str): Path to the opened netCDF file.
        f (netCDF4.Dataset): netCDF4 Dataset file handle.
        comm (Comm, optional): MPI communicator object. If None, just close the file directly.
            If communicator is specified, release the file lock after closing it.
    """
    f.close()
    if comm is not None and not comm.parallel_io:
        comm.release_file_lock(filename)

def nc_write_var(filename: str,
                 dim: Dict[str,Optional[int]],
                 varname: str,
                 dat: np.ndarray,
                 dtype: Optional[str]=None,
                 recno: Optional[Dict[str,int]]=None,
                 attr: Optional[Dict]=None,
                 comm: Optional[Comm]=None) -> None:
    """
    Write a variable to a netCDF file.

    Args:
        filename (str): Path to the output netCDF file.
        dim (dict): Dictionary {dimension name (str): dimension size (int)} of each dimension.
            The dimension size can be None if it is `unlimited` dimension (can append more records afterwards)
        varname (str): Name of the output variable. Variable groups are supported, use :code:`'group/varname'` as varname.
        dat (np.ndarray): Data for output, number of its dimensions must match :code:`dim` (excluding unlimited dimensions).
        dtype (str, optional): Data type to convert the input data to.
        recno (dict, optional): Dictionary {dimension name (str): record number (int)}, indicating the index in unlimited dimensions
            for the data to be written to. Each unlimited dimension defined in :code:`dim` should have
            a corresponding :code:`recno` entry.
        attr (dict, optional): Dictionary {name of attribute (str): value (str)}, additional attributes to be added.
        comm (Comm, optional): MPI communicator object, handling parallel I/O and make sure thread-safe writing of data.
    """
    f = nc_open(filename, 'a', comm)

    varname_parts = varname.split('/')
    group_path, varname = '/'.join(varname_parts[:-1]), varname_parts[-1]

    group = f
    if group_path:
        for part in group_path.split('/'):
            if part not in group.groups:
                group.createGroup(part)
            group = group.groups[part]

    if dtype is None:
        if isinstance(dat, np.ndarray):
            dtype = dat.dtype
        else:
            dtype = type(dat)

    ndim = len(dim)
    s = ()  ##slice for each dimension
    d = 0
    for i, name in enumerate(dim):
        if dim[name] is None:
            assert(name in recno), "unlimited dimension "+name+" doesn't have a recno"
            s += (recno[name],)
        else:
            s += (slice(None),)
            assert(dat.shape[d] == dim[name]), "dat size for dimension "+name+" mismatch with dim["+name+"]={}".format(dim[name])
            d += 1

        if name in group.dimensions:
            if dim[name] is not None:
                assert(group.dimensions[name].size==dim[name]), "dimension "+name+" size ({}) mismatch with file ({})".format(dim[name], group.dimensions[name].size)
            else:
                if not group.dimensions[name].isunlimited():
                    assert(recno[name] < group.dimensions[name].size), "recno for dimension "+name+" exceeds file size"
        else:
            group.createDimension(name, size=dim[name])

    if varname not in group.variables:
        group.createVariable(varname, dtype, tuple(dim.keys())) # type: ignore

    if isinstance(dat, np.ndarray):
        dat = dat.astype(dtype)
    else:
        dat = dtype(dat)

    group[varname][s] = dat  ##write dat to file

    if attr is not None:
        for akey in attr:
            group[varname].setncattr(akey, attr[akey])

    nc_close(filename, f, comm)

def nc_read_var(filename: str, varname: str, comm: Optional[Comm]=None) -> np.ndarray:
    """
    Read a variable from a netCDF file.

    This function by default reads the entire variable, if you only want a slice, it is more efficient to use
    netCDF4.Dataset handle directly.

    Args:
        filename (str): Path to the netCDF file for reading.
        varname (str): Name of the variable to read.
        comm (Comm, optional): MPI communicator object.

    Returns:
        np.ndarray: Variable read from the file.
    """
    f = nc_open(filename, 'r', comm)

    varname_parts = varname.split('/')
    group_path, varname = '/'.join(varname_parts[:-1]), varname_parts[-1]

    group = f
    if group_path:
        for part in group_path.split('/'):
            if part not in group.groups:
                group.createGroup(part)
            group = group.groups[part]

    assert varname in group.variables, f"variable '{varname}' is not defined in {filename}"

    dat = group[varname][...]
    dat_out = dat.data
    dat_out[dat.mask] = np.nan

    nc_close(filename, f, comm)

    return dat_out



from datetime import datetime
import os
import typing
import threading

import netCDF4 # type: ignore
import numpy as np
import pyproj # type: ignore

from NEDAS.grid import Grid
from NEDAS.models.nextsim.dg.perturb import gen_perturb, apply_perturb

_proj:pyproj.Proj = pyproj.Proj(proj='stere', a=6378273, b=6356889.448910593, lat_0=90., lon_0=-45., lat_ts=60.)

thread_lock = threading.Lock()
def read_var(fname:str, varnames:list[str]) -> np.ma.MaskedArray:
    """reading a variable from a netcdf file

    Parameters
    ----------
    fname : str
        forcing file name
    varnames : list[str]
        list of variable names
    itime : int
        time index in the forcing file
    """
    data: list[np.ndarray] = []
    with thread_lock:
        with netCDF4.Dataset(fname, 'r') as f:
            # read the variable
            for vname in varnames:
                data.append(f[vname][:])
    return np.ma.array(data)


def write_var(fname:str, varnames: list[str], data: np.ndarray) -> None:
    """Write the perturbed variable back to the forcing file

    Parameters
    ----------
    fname : str
        forcing file name
    varnames : list[str]
        list of variable names
    itime : int
        time index in the forcing file
    """
    # We assume all variables in the forcing file exists
    assert os.path.exists(fname), f'{fname} does not exist; Please copy the forcing file to the correct path first.'
    with thread_lock:
        with netCDF4.Dataset(fname, 'r+') as f:
            for i, vname in enumerate(varnames):
                f[vname][:] = data[i]
            f.sync()


def get_restart_filename(file_options:dict, i_ens: int, time: datetime) -> str:
    # get the restart file name
    fname: str
    try:
        fname = file_options['format'].format(i=i_ens, time=time.strftime(file_options['time_format']))
    except KeyError:
        try:
            fname = file_options['format'].format(i=i_ens, time=time.strftime(file_options['time_format']))
        except KeyError:
            print ('Currently, we only supports keyword of 1. "time_format",'
                   '2. "time_format"+"i".'
                   'See the example yaml file for more information. '
                   'Modified the code if you have other requirements.')
    return fname


def perturb_restart(restart_options:dict, file_options:dict, debug=False) -> None:
    """perturb the initial conditions in restart files

    Parameters
    ----------
    restart_options : dict
        perturbation options under the section of `perturb` from the yaml file
    file_options : dict
        This dictionary is constructed before the function is called.
        It contains the following keys
        - fname : str
            forcing file name.
            This has to be the file that will be perturbed, e.g. in the ensemble directory.
            This is usually derived before it is called.
        - lon_name: str
            name of the longitude variable in the forcing file.
            This is obtained from the files/restart section of the model configuration file.
        - lat_name: str
            name of the latitude variable in the forcing file
            This is obtained from the files/restart section of the model configuration file.
    """

    # perturbation arrays
    pert: np.ndarray[typing.Any, np.dtype[np.float64]]
    # get the restart file name
    fname:str = file_options['fname']
    # get grid object
    with thread_lock:
        with netCDF4.Dataset(fname, 'r') as f:
            lon = f[file_options['lon_name']][:]
            lat = f[file_options['lat_name']][:]
    x, y = _proj(lon, lat)
    grid:Grid = Grid(_proj, x, y)

    # get options for perturbing the forcing variables
    options = restart_options['variables']
    for i, varname in enumerate(options['names']):
        if typing.TYPE_CHECKING:
            assert type(varname) == str, 'variable name must be a string'

        # convert the horizontal correlation length scale to grid points
        hcorr:int = np.rint(float(options['hcorr'][i])/grid.dx)
        # do perturbation
        pert = gen_perturb(grid, options['type'][i], float(options['amp'][i]), hcorr)

        # apply perturbations to the variable data
        # in the case of vector fields, we need to split the variable name
        varname_list: list[str] = varname.split(';')
        # read the variable data from forcing file
        data: np.ndarray = read_var(fname, varname_list)
        # apply perturbations to the variable data
        data = apply_perturb(grid, data, pert, options['type'][i])
        # apply lower and upper bounds
        lb = float(options['lower_bounds'][i]) if options['lower_bounds'][i] != 'None' else -np.inf
        ub = float(options['upper_bounds'][i]) if options['upper_bounds'][i] != 'None' else np.inf
        data = np.minimum(np.maximum(data, lb), ub)
        # write the perturbed variable back to the forcing file
        write_var(fname, varname_list, data)



import os
import numpy as np
from NEDAS.utils.conversion import dt1h

def bool_str(value):
    if value:
        return '.true.'
    else:
        return '.false.'

def namelist(m, time, forecast_period, run_dir='.'):

    next_time = time + forecast_period * dt1h

    ###wps namelist.wps
    """
    nmlstr = "&share\n"
    nmlstr += "wrf_core = 'ARW',"
    nmlstr += "max_dom = 2,
    nmlstr += "start_date = '2000-01-24_12:00:00','2000-01-24_12:00:00',
    nmlstr += "end_date   = '2000-01-25_12:00:00','2000-01-24_12:00:00',
    nmlstr += "interval_seconds = 21600
    nmlstr += "active_grid = .true., .true.,
    nmlstr += "subgrid_ratio_x = 1
    nmlstr += "subgrid_ratio_y = 1
    nmlstr += "io_form_geogrid = 2,
    nmlstr += "opt_output_from_geogrid_path = './',
    nmlstr += "debug_level = 0
    nmlstr += "/\n\n"

    nmlstr += "&geogrid\n"
    nmlstr += "parent_id         =   1,   1,
    nmlstr += "parent_grid_ratio =   1,   3,
    nmlstr += "i_parent_start    =   1,  31,
    nmlstr += "j_parent_start    =   1,  17,
    nmlstr += "s_we              =   1,   1,
    nmlstr += "e_we              =  74, 112,
    nmlstr += "s_sn              =   1,   1,
    nmlstr += "e_sn              =  61,  97,
    nmlstr += "geog_data_res = 'default','default',
    nmlstr += "dx        = 30000,
    nmlstr += "dy        = 30000,
    nmlstr += "map_proj  = 'lambert',
    nmlstr += "ref_lat   =  34.83,
    nmlstr += "ref_lon   = -81.03,
    nmlstr += "ref_x     =  37.0,
    nmlstr += "ref_y     =  30.5,
    nmlstr += "truelat1  =  30.0,
    nmlstr += "truelat2  =  60.0,
    nmlstr += "stand_lon = -98.0,
    nmlstr += "geog_data_path = '/glade/work/wrfhelp/WPS_GEOG/'
    nmlstr += "opt_geogrid_tbl_path = 'geogrid/'
    nmlstr += "/
    nmlstr += "geog_data_res     = 'modis_lakes+10m','modis_lakes+2m',
    nmlstr += "geog_data_res     = 'usgs_lakes+10m','usgs_lakes+2m',

    nmlstr += "&ungrib
    nmlstr += "out_format = 'WPS',
    nmlstr += "prefix     = 'FILE',
    nmlstr += "ec_rec_len = 26214508,
    nmlstr += "pmin = 100.
    nmlstr += "/

    nmlstr += "&metgrid
    nmlstr += "fg_name         = 'FILE'
    nmlstr += "constants_name  = './TAVGSFC'
    nmlstr += "io_form_metgrid = 2, 
    nmlstr += "opt_output_from_metgrid_path = './',
    nmlstr += "opt_metgrid_tbl_path         = 'metgrid/',
    nmlstr += "process_only_bdy = 5,
    nmlstr += "/

    nmlstr += "&mod_levs
    nmlstr += "press_pa = 201300 , 200100 , 100000 , 
    nmlstr += "/\n\n"

    with open(os.path.join(run_dir, 'namelist.wps'), 'wt') as f:
        f.write(nmlstr)


    ##wrf namelist.input
    nmlstr = f"&time_control\n"
    nmlstr += f"run_days         = {int(forecast_period/24)},\n"
    nmlstr += f"run_hours        = {int(forecast_period%24)},\n"
    nmlstr += f"run_minutes      = {int(np.round(forecast_period - int(forecast_period)))},\n"
    nmlstr += f"run_seconds      = 0,\n"
    nmlstr += f"start_year       = 2019, 2019,
    nmlstr += f"start_month      = 09,   09, 
    nmlstr += f"start_day        = 04,   04,
    nmlstr += f"start_hour       = 12,   12,
    nmlstr += f"end_year         = 2019, 2019,
    nmlstr += f"end_month        = 09,   09,
    nmlstr += f"end_day          = 06,   06,
    nmlstr += f"end_hour         = 00,   00,
    nmlstr += f"interval_seconds = 10800
    nmlstr += f"input_from_file  = .true.,.true.,\n"
    nmlstr += f"history_interval = 60,  60,\n"
    nmlstr += f"frames_per_outfile = 1, 1,\n"
    nmlstr += f"restart            = .false.,\n"
    nmlstr += f"restart_interval   = 7200,\n"
    nmlstr += f"io_form_history    = 2,\n"
    nmlstr += f"io_form_restart    = 2,\n"
    nmlstr += f"io_form_input      = 2,\n"
    nmlstr += f"io_form_boundary   = 2,\n"
    nmlstr += f"/\n"

    nmlstr += "&domains
    nmlstr += "time_step                           = 90,
    nmlstr += "time_step_fract_num                 = 0,
    nmlstr += "time_step_fract_den                 = 1,
    nmlstr += "max_dom                             = 2,
    nmlstr += "e_we                                = 150,    220,
    nmlstr += "e_sn                                = 130,    214,
    nmlstr += "e_vert                              = 45,     45,
    nmlstr += "dzstretch_s                         = 1.1
    nmlstr += "p_top_requested                     = 5000,
    nmlstr += "num_metgrid_levels                  = 34,
    nmlstr += "num_metgrid_soil_levels             = 4,
    nmlstr += "dx                                  = 15000,
    nmlstr += "dy                                  = 15000,
    nmlstr += "grid_id                             = 1,     2,
    nmlstr += "parent_id                           = 0,     1,
    nmlstr += "i_parent_start                      = 1,     53,
    nmlstr += "j_parent_start                      = 1,     25,
    nmlstr += "parent_grid_ratio                   = 1,     3,
    nmlstr += "parent_time_step_ratio              = 1,     3,
    nmlstr += "feedback                            = 1,
    nmlstr += "smooth_option                       = 0
    nmlstr += "/

    nmlstr += "&physics
    nmlstr += "physics_suite                       = 'CONUS'
    nmlstr += "mp_physics                          = -1,    -1,
    nmlstr += "cu_physics                          = -1,    -1,
    nmlstr += "ra_lw_physics                       = -1,    -1,
    nmlstr += "ra_sw_physics                       = -1,    -1,
    nmlstr += "bl_pbl_physics                      = -1,    -1,
    nmlstr += "sf_sfclay_physics                   = -1,    -1,
    nmlstr += "sf_surface_physics                  = -1,    -1,
    nmlstr += "radt                                = 15,    15,
    nmlstr += "bldt                                = 0,     0,
    nmlstr += "cudt                                = 0,     0,
    nmlstr += "icloud                              = 1,
    nmlstr += "num_land_cat                        = 21,
    nmlstr += "sf_urban_physics                    = 0,     0,
    nmlstr += "fractional_seaice                   = 1,
    nmlstr += "/

    nmlstr += "&fdda
    nmlstr += "/

    nmlstr += "&dynamics
    nmlstr += "hybrid_opt                          = 2, 
    nmlstr += "w_damping                           = 0,
    nmlstr += "diff_opt                            = 2,      2,
    nmlstr += "km_opt                              = 4,      4,
    nmlstr += "diff_6th_opt                        = 0,      0,
    nmlstr += "iff_6th_factor                     = 0.12,   0.12,
    nmlstr += "base_temp                           = 290.
    nmlstr += "damp_opt                            = 3,
    nmlstr += "zdamp                               = 5000.,  5000.,
    nmlstr += "dampcoef                            = 0.2,    0.2,
    nmlstr += "khdif                               = 0,      0,
    nmlstr += "kvdif                               = 0,      0,
    nmlstr += "non_hydrostatic                     = .true., .true.,
    nmlstr += "moist_adv_opt                       = 1,      1,
    nmlstr += "scalar_adv_opt                      = 1,      1,
    nmlstr += "gwd_opt                             = 1,      0,
    nmlstr += "/

    nmlstr += "&bdy_control
    nmlstr += "spec_bdy_width                      = 5,
    nmlstr += "specified                           = .true.
    nmlstr += "/

    nmlstr += "&grib2
    nmlstr += "/

    nmlstr += "&namelist_quilt
    nmlstr += "nio_tasks_per_group = 0,
    nmlstr += "nio_groups = 1,
    nmlstr += "/

    with open(os.path.join(run_dir, 'namelist.input'), 'wt') as f:
        f.write(nmlstr)
    """


import numpy as np
import os
import glob
from datetime import datetime, timedelta, timezone
import netCDF4
import pyproj
from NEDAS.grid import Grid
from NEDAS.datasets import Dataset

class OsisafSeaIceConcObs(Dataset):

    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        self.variables = {'seaice_conc': {'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':1}, }

        proj = pyproj.Proj(self.proj)
        x, y = np.meshgrid(np.arange(self.xstart, self.xend, self.dx), np.arange(self.ystart, self.yend, self.dy))
        self.grid = Grid(proj, x, y)

    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        path = kwargs['path']
        time = kwargs['time']
        name = kwargs['name']
        obs_window_min = kwargs['obs_window_min']
        obs_window_max = kwargs['obs_window_max']

        if time is None:
            search = os.path.join(path, '????_'+self.proj_name, 'ice_conc_'+self.proj_name+'-100_multi_????????????.nc')
            file_list = glob.glob(search)

        else:
            if obs_window_min is not None and obs_window_max is not None:
                d_range = np.arange(obs_window_min, obs_window_max)
            else:
                d_range = [0]

            file_list = []
            for d in d_range:
                t = time + d * timedelta(hours=1)
                tstr = t.strftime('%Y%m%d%H%M')
                search = os.path.join(path, t.strftime('%Y')+'_'+self.proj_name, 'ice_conc_'+self.proj_name+'-100_multi_'+tstr+'.nc')
                for result in glob.glob(search):
                    if result not in file_list:
                        file_list.append(result)

        assert len(file_list)>0, 'no matching files found: '+search
        return file_list

    def random_network(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)

        if kwargs['nobs'] is None:
            nobs = 1000
        else:
            nobs = kwargs['nobs']

        grid = kwargs['grid']

        obs_x = []
        obs_y = []
        while len(obs_x) < nobs:
            y = np.random.uniform(grid.ymin, grid.ymax)
            x = np.random.uniform(grid.xmin, grid.xmax)
            valid = grid.interp(grid.mask.astype(int), x, y)
            if valid == 0:
                obs_x.append(x)
                obs_y.append(y)

        obs_x = np.array(obs_x)
        obs_y = np.array(obs_y)
        obs_z = np.zeros(nobs)

        obs_seq = {'obs': np.full(nobs, np.nan),
                't': np.full(nobs, kwargs['time']),
                'z': obs_z,
                'y': obs_y,
                'x': obs_x,
                'err_std': np.ones(nobs) * kwargs['err']['std']
                }
        return obs_seq

    def read_obs(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        grid = kwargs['grid']
        mask = kwargs['mask']

        obs_seq = {'obs':[], 't':[], 'z':[], 'y':[], 'x':[], 'err_std':[], }

        for fname in self.filename(**kwargs):

            ##read the data file
            f = netCDF4.Dataset(fname)

            lat = f['lat'][...].data.flatten()
            lon = f['lon'][...].data.flatten()
            x_, y_ = grid.proj(lon, lat)
            mask_ = grid.interp(mask.astype(int), x_, y_)

            ntime = f.dimensions['time'].size
            for n in range(ntime):
                t = f['time'][n].data * timedelta(seconds=1) + datetime(1978, 1, 1, tzinfo=timezone.utc)
                qc_flag = f['status_flag'][n,...].data.flatten()
                obs = f['ice_conc'][n,...].data.flatten()
                if 'total_uncertainty' in f.variables:
                    obs_err = f['total_uncertainty'][n,...].data.flatten()
                else:
                    obs_err = 0.3 * np.ones(obs.shape)  ##default conc err 0.3

                for p in range(obs.size):
                    if qc_flag[p] < 0 or qc_flag[p] > 10:
                        continue
                    if obs_err[p] <= 0:
                        continue

                    if x_[p] < grid.xmin or x_[p] > grid.xmax or y_[p] < grid.ymin or y_[p] > grid.ymax:
                        continue

                    if mask_[p] > 0:
                        continue

                    obs_value = obs[p] * 0.01   ##convert percent to 0-1

                    #obs_err_var = (obs_err[p]*0.01)**2  ##uncertainty from dataset
                    obs_err_var = 1. + (0.5 - np.abs(0.5-obs_value))**2  ##adaptive error used in topaz

                    ##assignn to obs_seq
                    obs_seq['obs'].append(obs_value)
                    obs_seq['err_std'].append(np.sqrt(obs_err_var))
                    obs_seq['t'].append(t)
                    obs_seq['z'].append(0)
                    obs_seq['y'].append(y_[p])
                    obs_seq['x'].append(x_[p])

            f.close()

        for key in obs_seq.keys():
            obs_seq[key] = np.array(obs_seq[key])

        return obs_seq




from NEDAS.config import Config
from NEDAS.assim_tools.state import State
from .base import Obs

def get_obs(c: Config, state: State) -> Obs:
    # obs_type = [r['type'] for r in c.obs_def]
    # module = importlib.import_module('NEDAS.assim_tools.obs.'+obs_type)
    # ObsClass = getattr(module, registry[obs_type])
    # return ObsClass(c, state)
    return Obs(c, state)
